#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import typing

from sane_finances.communication.downloader import DownloadStringResult, Downloader
from sane_finances.sources.base import (
    InstrumentStringDataDownloader,
    InstrumentValuesHistoryParser, InstrumentInfoParser,
    InstrumentInfoProvider, InstrumentValueProvider,
    InstrumentHistoryDownloadParameters, InstrumentInfo, InstrumentValue, InstrumentExporterFactory,
    DownloadParametersFactory, DynamicEnumTypeManager, ApiActualityChecker, DownloadParameterValuesStorage,
    InstrumentsInfoExporter, InstrumentHistoryValuesExporter)


class FakeInstrumentInfoProvider(InstrumentInfoProvider):

    def __init__(self, value: InstrumentInfo):
        self.value = value

    @property
    def instrument_info(self) -> InstrumentInfo:
        return self.value


class FakeInstrumentValueProvider(InstrumentValueProvider):

    def __init__(self, value: InstrumentValue):
        self.value = value

    def get_instrument_value(self, tzinfo: typing.Optional[datetime.timezone]) -> InstrumentValue:
        return self.value


class FakeInstrumentHistoryDownloadParameters(InstrumentHistoryDownloadParameters):

    def clone_with_instrument_info_parameters(
            self,
            info_download_parameters: typing.Any,
            instrument_info: typing.Optional[InstrumentInfoProvider]) -> InstrumentHistoryDownloadParameters:
        return FakeInstrumentHistoryDownloadParameters()


class FakeInstrumentStringDataDownloader(InstrumentStringDataDownloader):

    def __init__(self, fake_info_string: str, fake_history_string: str):
        self.fake_info_string = fake_info_string
        self.fake_history_string = fake_history_string

        self.download_instruments_info_string_results: typing.List[DownloadStringResult] = []
        self.download_instrument_history_string_results: typing.List[DownloadStringResult] = []

        self.download_instruments_info_string_counter = 0
        self.download_instrument_history_string_counter = 0

    def adjust_download_instrument_history_parameters(
            self,
            parameters: InstrumentHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Tuple[InstrumentHistoryDownloadParameters, datetime.datetime, datetime.datetime]:
        # by default DO NOT adjust anything for testing simplicity !!!
        return parameters, moment_from, moment_to

    def paginate_download_instrument_history_parameters(
            self,
            parameters: InstrumentHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Iterable[typing.Tuple[InstrumentHistoryDownloadParameters, datetime.datetime, datetime.datetime]]:
        # by default DO NOT paginate anything for testing simplicity !!!
        yield parameters, moment_from, moment_to

    def download_instruments_info_string(self, parameters) -> DownloadStringResult:

        self.download_instruments_info_string_counter += 1
        result = DownloadStringResult(self.fake_info_string)
        self.download_instruments_info_string_results.append(result)
        return result

    def download_instrument_history_string(
            self,
            parameters: InstrumentHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime) -> DownloadStringResult:

        self.download_instrument_history_string_counter += 1
        result = DownloadStringResult(self.fake_history_string)
        self.download_instrument_history_string_results.append(result)
        return result


class FakeInstrumentValuesHistoryParser(InstrumentValuesHistoryParser):
    fake_result: typing.Iterable[InstrumentValueProvider]

    def __init__(self, expected_raw_text: str):
        self.expected_raw_text = expected_raw_text
        self.parse_exception = None

        self.parse_counter = 0

    def parse(
            self,
            raw_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[InstrumentValueProvider]:
        if raw_text != self.expected_raw_text:
            raise ValueError("Not expected 'raw_text'")

        if self.parse_exception is not None:
            raise self.parse_exception

        self.parse_counter += 1
        yield from self.fake_result


class FakeInstrumentInfoParser(InstrumentInfoParser):
    fake_result: typing.Iterable[InstrumentInfoProvider]

    def __init__(self, expected_raw_text: str):
        self.expected_raw_text = expected_raw_text
        self.parse_exception = None

        self.parse_counter = 0

    def parse(self, raw_text: str) -> typing.Iterable[InstrumentInfoProvider]:
        if raw_text != self.expected_raw_text:
            raise ValueError("Not expected 'raw_text'")

        if self.parse_exception is not None:
            raise self.parse_exception

        self.parse_counter += 1
        return self.fake_result


class FakeInstrumentExporterFactory(InstrumentExporterFactory):
    history_values_exporter: InstrumentHistoryValuesExporter
    info_exporter: InstrumentsInfoExporter
    download_parameter_values_storage: DownloadParameterValuesStorage
    api_actuality_checker: ApiActualityChecker

    def __init__(
            self,
            dynamic_enum_type_manager: DynamicEnumTypeManager = None,
            download_parameters_factory: DownloadParametersFactory = None):
        self._dynamic_enum_type_manager = dynamic_enum_type_manager
        self._download_parameters_factory = download_parameters_factory

    def create_history_values_exporter(self, downloader: Downloader) -> InstrumentHistoryValuesExporter:
        return self.history_values_exporter

    def create_info_exporter(self, downloader: Downloader) -> InstrumentsInfoExporter:
        return self.info_exporter

    def create_download_parameter_values_storage(self, downloader: Downloader) -> DownloadParameterValuesStorage:
        return self.download_parameter_values_storage

    def create_api_actuality_checker(self, downloader: Downloader) -> ApiActualityChecker:
        return self.api_actuality_checker

    @property
    def dynamic_enum_type_manager(self) -> DynamicEnumTypeManager:
        return self._dynamic_enum_type_manager

    @property
    def download_parameters_factory(self) -> DownloadParametersFactory:
        return self._download_parameters_factory



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from sane_finances.sources.base import ParseError
from sane_finances.sources.solactive.v2018.meta import IndexInfo
from sane_finances.sources.solactive.v2018.parsers import SolactiveJsonParser, SolactiveIndexInfoParser


class TestSolactiveJsonParser(unittest.TestCase):

    def setUp(self) -> None:
        self.parser = SolactiveJsonParser()

    def check_parse_raise(self, invalid_json: str):
        with self.assertRaises(ParseError):
            _ = list(self.parser.parse(invalid_json, tzinfo=None))

    def test_parse_Success(self):
        valid_json = '[{"indexId":"DE000SLA4YD9","timestamp":1147046400000,"value":466.44}]'

        _ = list(self.parser.parse(valid_json, tzinfo=None))

    def test_parse_raisesWhenNoData(self):
        invalid_json = ''

        self.check_parse_raise(invalid_json)

    def test_parse_raisesWhenEmptyList(self):
        invalid_json = '[]'

        self.check_parse_raise(invalid_json)

    def test_parse_raisesWhenNotList(self):
        invalid_json = '{}'

        self.check_parse_raise(invalid_json)

    def test_parse_raisesWhenNotDict(self):
        invalid_json = '[[]]'

        self.check_parse_raise(invalid_json)

    def test_parse_raisesWhenNoIndexIDField(self):
        invalid_json = '[{"indexId__":"DE000SLA4YD9","timestamp":1147046400000,"value":466.44}]'

        self.check_parse_raise(invalid_json)

    def test_parse_raisesWhenNoTimestampField(self):
        invalid_json = '[{"indexId":"DE000SLA4YD9","timestamp__":1147046400000,"value":466.44}]'

        self.check_parse_raise(invalid_json)

    def test_parse_raisesWhenNoValueField(self):
        invalid_json = '[{"indexId":"DE000SLA4YD9","timestamp":1147046400000,"value__":466.44}]'

        self.check_parse_raise(invalid_json)


class TestSolactiveIndexInfoParser(unittest.TestCase):
    
    def test_parse_success(self):
        expected_result = [IndexInfo(name='11USA Index', isin='DE000SLA3E30'),
                           IndexInfo(name='ABCD', isin='DETEST104000')]
        
        first_name, first_isin = expected_result[0].name, expected_result[0].isin
        second_name, second_isin = expected_result[1].name, expected_result[1].isin
        
        html = f"""
<!doctype html>
<html lang="en-US">
  <head>
  <meta charset="utf-8">
  <meta http-equiv="x-ua-compatible" content="ie=edge">
  <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
  <title>Solactive | Indices</title>

</script>  <script src='https://www.google.com/recaptcha/api.js?hl=en'></script>
</head>


  <body class="page-template page-template-template-indices template-indices page page-id-11 page-parent ">
        <header class="header navbar navbar-inverse">

</header>
        <table class="groups table table-hover datatable-indices dataTable no-footer" data-pagelength="15">
        <thead>
            <tr>
                <th class="name">Name</th>
                <th class="isin">ISIN</th>
                <th class="level">Level</th>
            </tr>
        </thead>
        <tbody>
                            <tr>
                    <td class="name">
                        <a href="?index=DE000SLA3E30">{first_name}</a>
                    </td>
                    <td class="isin">{first_isin}</td>
                    <td class="level">603.12</td>
                </tr>
                            <tr>
                    <td class="name">
                        <a href="?index=DETEST104000">{second_name}</a>
                    </td>
                    <td class="isin">{second_isin}</td>
                    <td class="level">101.18</td>
                </tr>
                    </tbody>
    </table>

    </div>
</section>


        </main>
<script type='text/javascript' src='https://www.solactive.com/wp-includes/js/wp-embed.min.js?ver=5.2.5'></script>
  </body>
</html>
        """
        
        parser = SolactiveIndexInfoParser()
        
        result = list(parser.parse(html))
        
        self.assertSequenceEqual(result, expected_result)
    
    def test_parse_raisesWithoutName(self):
        html = """
        <table class="groups table table-hover datatable-indices dataTable no-footer" data-pagelength="15">
        <thead>
            <tr>
                <th class="name">Name</th>
                <th class="isin">ISIN</th>
                <th class="level">Level</th>
            </tr>
        </thead>
        <tbody>
                            <tr>
                    <td class="">
                        <a href="?index=DE000SLA3E30">11USA Index</a>
                    </td>
                    <td class="isin">DE000SLA3E30</td>
                    <td class="level">603.12</td>
                </tr>
                    </tbody>
    </table>
        """

        parser = SolactiveIndexInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(html))

    def test_parse_raisesWithEmptyName(self):
        html = """
        <table class="groups table table-hover datatable-indices dataTable no-footer" data-pagelength="15">
        <thead>
            <tr>
                <th class="name">Name</th>
                <th class="isin">ISIN</th>
                <th class="level">Level</th>
            </tr>
        </thead>
        <tbody>
                            <tr>
                    <td class="name">
                    </td>
                    <td class="isin">DE000SLA3E30</td>
                    <td class="level">603.12</td>
                </tr>
                    </tbody>
    </table>
        """

        parser = SolactiveIndexInfoParser()

        with self.assertRaises(ParseError):
            list(parser.parse(html))

    def test_parse_raisesWithoutIsin(self):
        html = """
        <table class="groups table table-hover datatable-indices dataTable no-footer" data-pagelength="15">
        <thead>
            <tr>
                <th class="name">Name</th>
                <th class="isin">ISIN</th>
                <th class="level">Level</th>
            </tr>
        </thead>
        <tbody>
                            <tr>
                    <td class="name">
                        <a href="?index=DE000SLA3E30">11USA Index</a>
                    </td>
                    <td class="">DE000SLA3E30</td>
                    <td class="level">603.12</td>
                </tr>
                    </tbody>
    </table>
        """

        parser = SolactiveIndexInfoParser()
        
        with self.assertRaises(ParseError):
            list(parser.parse(html))
    
    def test_parse_raisesEmptyString(self):
        html = ""

        parser = SolactiveIndexInfoParser()
        
        with self.assertRaises(ParseError):
            list(parser.parse(html))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Metadata for export index data from solactive.com
"""

import decimal
import datetime
import enum
import typing
import dataclasses

from ...base import (
    InstrumentValue, InstrumentInfo, InstrumentValueProvider, InstrumentInfoProvider,
    InstrumentHistoryDownloadParameters, DownloadParametersFactory)
from ...inspection import InstrumentInfoParameter
from ....annotations import LEGACY_ANNOTATIONS

if LEGACY_ANNOTATIONS:  # pragma: no cover
    from ....annotations import Annotated
else:  # pragma: no cover
    from typing import Annotated


class IndexHistoryTypes(enum.Enum):
    """ History type (period).
    """
    INTRADAY = 'intraday'
    MAX = 'max'


class FileExtensions(enum.Enum):
    """ File extension.
    """
    JSON = '.json'


class FieldNames(enum.Enum):
    """ Field name in JSON.
    """
    INDEX_ID = 'indexId'
    TIMESTAMP = 'timestamp'
    VALUE = 'value'


@dataclasses.dataclass
class IndexValue(InstrumentValueProvider):
    """ Container for index history item.
    """
    index_id: str
    moment: datetime.datetime
    value: decimal.Decimal

    def __init__(self, *, index_id: str, moment: datetime.datetime, value: decimal.Decimal):
        """ Initialize index value.

        :param index_id: Index ID
        :param moment: Moment.
        :param value: Value of index.
        """
        if not isinstance(moment, datetime.datetime):
            raise TypeError("'moment' is not datetime")

        self.index_id = str(index_id)
        self.moment = moment
        self.value = decimal.Decimal(value)

    def get_instrument_value(self, tzinfo: typing.Optional[datetime.timezone]) -> InstrumentValue:
        moment = self.moment.astimezone(tzinfo)
        return InstrumentValue(value=self.value, moment=moment)


@dataclasses.dataclass
class IndexInfo(InstrumentInfoProvider):
    """ Container for index information.
    """
    isin: str
    name: str

    def __init__(self, *, isin: str, name: str):
        self.isin = str(isin)
        self.name = str(name)

    def __str__(self):
        return (f"Splactive index ("
                f"isin={self.isin}, "
                f"name={self.name})")

    @property
    def instrument_info(self) -> InstrumentInfo:
        return InstrumentInfo(code=self.isin, name=self.name)


class SolactiveIndexesInfoDownloadParameters(typing.NamedTuple):
    """ Container for ``SolactiveStringDataDownloader.download_instruments_info_string`` parameters.
    """

    @classmethod
    def safe_create(
            cls: typing.Type['SolactiveIndexesInfoDownloadParameters']) -> 'SolactiveIndexesInfoDownloadParameters':
        """ Create new instance of ``SolactiveIndexesInfoDownloadParameters`` with arguments check.

        :return: ``SolactiveIndexesInfoDownloadParameters`` instance.
        """
        return cls()


@dataclasses.dataclass
class SolactiveIndexHistoryDownloadParameters(InstrumentHistoryDownloadParameters):
    """ Container for ``SolactiveStringDataDownloader.download_instrument_history_string`` parameters.
    """
    isin: Annotated[str, InstrumentInfoParameter(instrument_identity=True)]

    def clone_with_instrument_info_parameters(
            self,
            info_download_parameters: typing.Optional[SolactiveIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> 'SolactiveIndexHistoryDownloadParameters':
        return SolactiveIndexHistoryDownloadParameters.generate_from(self, info_download_parameters, instrument_info)

    # noinspection PyUnusedLocal
    @classmethod
    def generate_from(
            cls: typing.Type['SolactiveIndexHistoryDownloadParameters'],
            history_download_parameters: typing.Optional['SolactiveIndexHistoryDownloadParameters'],
            info_download_parameters: typing.Optional[SolactiveIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> 'SolactiveIndexHistoryDownloadParameters':
        """ Create new history download parameters instance with data from its arguments.

        :param history_download_parameters: Optional instrument history download parameters for cloning.
        :param info_download_parameters: Optional instrument info download parameters for cloning.
        :param instrument_info: Optional instrument info for cloning.
        :return: Cloned history download parameters instance (self) with replacing some attributes from
            `info_download_parameters` and `instrument_info`.
        """
        return cls(
            isin=((None if history_download_parameters is None else history_download_parameters.isin)
                  if instrument_info is None
                  else instrument_info.isin)
        )

    @classmethod
    def safe_create(
            cls: typing.Type['SolactiveIndexHistoryDownloadParameters'],
            *,
            isin: str) -> 'SolactiveIndexHistoryDownloadParameters':
        """ Create new instance of ``SolactiveIndexHistoryDownloadParameters`` with arguments check.

        :param isin: ISIN.
        :return: ``SolactiveIndexHistoryDownloadParameters`` instance.
        """
        return cls(isin=str(isin))


class SolactiveDownloadParametersFactory(DownloadParametersFactory):
    """ Download parameters factories and generators for Solactive.
    """

    @property
    def download_history_parameters_class(self) -> typing.Type[SolactiveIndexHistoryDownloadParameters]:
        return SolactiveIndexHistoryDownloadParameters

    @property
    def download_history_parameters_factory(self) -> typing.Callable[..., SolactiveIndexHistoryDownloadParameters]:
        return SolactiveIndexHistoryDownloadParameters.safe_create

    @property
    def download_info_parameters_class(self):
        return SolactiveIndexesInfoDownloadParameters

    @property
    def download_info_parameters_factory(self) -> typing.Callable[..., typing.Any]:
        return SolactiveIndexesInfoDownloadParameters.safe_create

    def generate_history_download_parameters_from(
            self,
            history_download_parameters: typing.Optional[SolactiveIndexHistoryDownloadParameters],
            info_download_parameters: typing.Optional[SolactiveIndexesInfoDownloadParameters],
            instrument_info: typing.Optional[IndexInfo]) -> SolactiveIndexHistoryDownloadParameters:
        return SolactiveIndexHistoryDownloadParameters.generate_from(
            history_download_parameters,
            info_download_parameters,
            instrument_info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Tools for download (export) currency rates data from cbr.ru
"""
import logging
import typing
import decimal
import datetime
import inspect

from .meta import (
    RateFrequencies, CbrDownloadParametersFactory,
    CbrCurrenciesInfoDownloadParameters, CbrCurrencyHistoryDownloadParameters)
from .parsers import CbrCurrencyHistoryXmlParser, CbrCurrencyInfoParser
from ...base import (
    ApiActualityChecker, InstrumentStringDataDownloader, ParseError,
    InstrumentExporterFactory, InstrumentHistoryValuesExporter, InstrumentsInfoExporter,
    DownloadParameterValuesStorage, CheckApiActualityError)
from ...generic import GenericInstrumentHistoryValuesExporter, GenericInstrumentsInfoExporter
from ....communication.downloader import Downloader, DownloadStringResult

logging.getLogger().addHandler(logging.NullHandler())


class CbrStringDataDownloader(InstrumentStringDataDownloader):
    """ Data downloader from cbr.ru.

    See http://www.cbr.ru/development/sxml/ for details.

    The list of all currencies can be obtained from
    https://www.cbr.ru/scripts/XML_val.asp?d=0
    and
    https://www.cbr.ru/scripts/XML_val.asp?d=1
    """

    CurrenciesListUrl = 'https://www.cbr.ru/scripts/XML_val.asp'
    RatesHistoryUrl = 'https://www.cbr.ru/scripts/XML_dynamic.asp'

    encoding = 'cp1251'
    query_date_format = '%d/%m/%Y'

    def __init__(self, downloader: Downloader):
        """ Initialize downloader.

        :param downloader: Used string downloader.
        """
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.downloader = downloader

        # headers for HTTP
        self.headers: typing.Dict[str, str] = {}

    def download_instrument_history_string(
            self,
            parameters: CbrCurrencyHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime) -> DownloadStringResult:

        date_from = moment_from.date()
        date_to = moment_to.date()
        if moment_to.time() != datetime.time.min:
            date_to += datetime.timedelta(days=1)

        return self.download_currency_history_string(parameters.currency_id, date_from, date_to)

    def download_instruments_info_string(
            self,
            parameters: CbrCurrenciesInfoDownloadParameters) -> DownloadStringResult:
        return self.download_currencies_info_string(parameters.rate_frequency)

    def download_currency_history_string(
            self,
            currency_id: str,
            date_from: datetime.date,
            date_to: datetime.date) -> DownloadStringResult:
        """ Downloads data for one currency as string.

        :param currency_id: Currency ID.
        :param date_from: Download interval beginning.
        :param date_to: Download interval ending.
        :return: Container with downloaded string.
        """
        params = [
            ('date_req1', date_from.strftime(self.query_date_format)),
            ('date_req2', date_to.strftime(self.query_date_format)),
            ('VAL_NM_RQ', currency_id)
        ]

        self.downloader.parameters = params
        self.downloader.headers = self.headers

        return self.downloader.download_string(self.RatesHistoryUrl, self.encoding)

    def download_currencies_info_string(self, rate_frequency: RateFrequencies) -> DownloadStringResult:
        """ Downloads the list of all available currencies by specified parameters.

        :param rate_frequency: ``RateFrequencies`` value.
        :return: Container with downloaded string.
        """
        self.downloader.headers = {}
        self.downloader.parameters = [
            ('d', rate_frequency.value)
        ]

        return self.downloader.download_string(self.CurrenciesListUrl, self.encoding)


class CbrApiActualityChecker(ApiActualityChecker):
    """ Verifies actuality and accessibility of REST API of cbr.ru.
    """

    _currencyToCheck = 'R01235'  # US Dollar
    _expectedFirstDate = datetime.date(2000, 1, 1)
    _expectedFirstNominal = 1
    _expectedFirstValue = decimal.Decimal('27')

    def __init__(self,
                 string_data_downloader: CbrStringDataDownloader,
                 history_xml_parser: CbrCurrencyHistoryXmlParser,
                 index_info_parser: CbrCurrencyInfoParser):
        """ Initialize checker.

        :param string_data_downloader: Used string data downloader.
        :param history_xml_parser: Used history xml parser.
        :param index_info_parser: Used index info parser.
        """
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.string_data_downloader = string_data_downloader
        self.history_xml_parser = history_xml_parser
        self.index_info_parser = index_info_parser

    def check(self):
        self.logger.info("Check actuality via currencies list")

        rate_frequency = RateFrequencies.DAILY
        history_date_from = self._expectedFirstDate
        history_date_to = self._expectedFirstDate + datetime.timedelta(days=10)

        xml_list_result = self.string_data_downloader.download_currencies_info_string(rate_frequency)
        self.logger.debug(f"Got XML data:\n{xml_list_result.downloaded_string}")

        try:
            info_list = list(self.index_info_parser.parse(xml_list_result.downloaded_string))
        except ParseError as ex:
            xml_list_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected currency info XML: {ex.message}") from ex
        except Exception:
            xml_list_result.set_correctness(False)
            raise

        if len(info_list) == 0:
            xml_list_result.set_correctness(False)
            raise CheckApiActualityError("Unexpected currency info list. No data")

        # here we know that check via index list was successful

        self.logger.info(f"Check actuality via currency {self._currencyToCheck!r}")

        xml_history_result = self.string_data_downloader.download_currency_history_string(
            self._currencyToCheck,
            history_date_from,
            history_date_to
        )

        try:
            history = list(self.history_xml_parser.parse(xml_history_result.downloaded_string, tzinfo=None))
        except ParseError as ex:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected currency history XML: {ex.message}") from ex
        except Exception:
            xml_history_result.set_correctness(False)
            raise

        history.sort(key=lambda v: v.date)

        first_value = history[0]  # for actuality checking one record is enough

        if first_value.currency_id != self._currencyToCheck:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected currency ID. "
                                         f"{first_value.currency_id!r} != {self._currencyToCheck!r}")

        if first_value.date != self._expectedFirstDate:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected first date. "
                                         f"{first_value.date} != {self._expectedFirstDate}")

        if first_value.nominal != self._expectedFirstNominal:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected first nominal. "
                                         f"{first_value.nominal} != {self._expectedFirstNominal}")

        if first_value.value != self._expectedFirstValue:
            xml_history_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected first value. "
                                         f"{first_value.value} != {self._expectedFirstValue}")

        self.logger.info("Actuality check was successful")


class CbrCurrencyDownloadParameterValuesStorage(DownloadParameterValuesStorage):
    """ Storage of currency download parameters.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self._special_handlers: typing.Dict[type, typing.Callable] = {
            RateFrequencies: self._get_rate_frequency_choices
        }

    def is_dynamic_enum_type(self, cls: type) -> bool:
        return False

    def get_all_managed_types(self) -> typing.Iterable[typing.Type]:
        return ()

    def get_dynamic_enum_key(self, instance):
        return None

    def get_parameter_type_choices(self, cls: type) \
            -> typing.Optional[
                typing.List[typing.Tuple[typing.Any, typing.Union[str, typing.List[typing.Tuple[typing.Any, str]]]]]
            ]:

        if not inspect.isclass(cls):
            return None

        if cls not in self._special_handlers:
            return None

        return self._special_handlers[cls]()

    @staticmethod
    def _get_rate_frequency_choices():
        rate_frequency: RateFrequencies

        return [(rate_frequency.value, rate_frequency.description)
                for rate_frequency
                in RateFrequencies]


class CbrCurrencyRatesExporterFactory(InstrumentExporterFactory):
    """ Factory class for create instances of Bank of Russia data exporter.
    """
    name: str = 'Bank of Russia Foreign Currency Market Lib. Version 1.4, 2016.'
    provider_site: str = 'http://www.cbr.ru/'
    api_url: str = 'http://www.cbr.ru/development/sxml/'

    def __init__(self):
        self._dynamic_enum_type_manager = None
        self._download_parameters_factory = None

    def create_history_values_exporter(self, downloader: Downloader) -> InstrumentHistoryValuesExporter:
        string_data_downloader = CbrStringDataDownloader(downloader)
        history_values_parser = CbrCurrencyHistoryXmlParser()

        return GenericInstrumentHistoryValuesExporter(string_data_downloader, history_values_parser)

    def create_info_exporter(self, downloader: Downloader) -> InstrumentsInfoExporter:
        string_data_downloader = CbrStringDataDownloader(downloader)
        info_parser = CbrCurrencyInfoParser()

        return GenericInstrumentsInfoExporter(string_data_downloader, info_parser)

    def create_download_parameter_values_storage(
            self,
            downloader: Downloader) -> CbrCurrencyDownloadParameterValuesStorage:
        return CbrCurrencyDownloadParameterValuesStorage()

    def create_api_actuality_checker(self, downloader: Downloader) -> CbrApiActualityChecker:
        string_data_downloader = CbrStringDataDownloader(downloader)
        history_values_parser = CbrCurrencyHistoryXmlParser()
        info_parser = CbrCurrencyInfoParser()

        return CbrApiActualityChecker(string_data_downloader, history_values_parser, info_parser)

    @property
    def dynamic_enum_type_manager(self) -> CbrCurrencyDownloadParameterValuesStorage:
        if self._dynamic_enum_type_manager is None:
            self._dynamic_enum_type_manager = CbrCurrencyDownloadParameterValuesStorage()

        return self._dynamic_enum_type_manager

    @property
    def download_parameters_factory(self) -> CbrDownloadParametersFactory:
        if self._download_parameters_factory is None:
            self._download_parameters_factory = CbrDownloadParametersFactory()

        return self._download_parameters_factory



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import unittest

from sane_finances.communication.cachers import ExpiryCalculator, DummyCacher, InMemoryCacher

from .fakes import FakeExpiryCalculator


class TestExpiryCalculator(unittest.TestCase):

    def test_is_expired_Success(self):
        yesterday = datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(days=1)
        tomorrow = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(days=1)
        calculator = ExpiryCalculator()

        self.assertTrue(calculator.is_expired(yesterday))
        self.assertFalse(calculator.is_expired(tomorrow))

    def test_get_revive_moment_Success(self):
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        calculator = ExpiryCalculator()

        revive_moment = calculator.get_revive_moment()

        self.assertGreaterEqual(revive_moment, now)

    def test_get_expiry_moment_Success(self):
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        delta = datetime.timedelta(days=1)
        expected_result = now + delta
        calculator = ExpiryCalculator()

        result = calculator.get_expiry_moment(delta, now)

        self.assertEqual(result, expected_result)

    def test_get_expiry_moment_ShiftFromNow(self):
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        delta = datetime.timedelta(days=1)
        expected_result = now + delta
        calculator = ExpiryCalculator()

        result = calculator.get_expiry_moment(delta)  # no start_from

        self.assertGreaterEqual(result, expected_result)


class TestDummyCacher(unittest.TestCase):

    def setUp(self):
        self.dummy_reviver_count = 0
        self.success_string = 'OK'

    def dummy_reviver(self):
        self.dummy_reviver_count += 1
        return self.success_string

    def test_retrieve_Success(self):
        cacher = DummyCacher()

        (got_from_cache, result) = cacher.retrieve('', [], {}, self.dummy_reviver)

        self.assertFalse(got_from_cache)
        self.assertEqual(result, self.success_string)
        self.assertEqual(self.dummy_reviver_count, 1)

    def test_drop_NeverDrop(self):
        cacher = DummyCacher()

        result = cacher.drop('', [], {})

        self.assertFalse(result)

    def test_has_NeverHas(self):
        cacher = DummyCacher()

        result = cacher.has('', [], {})

        self.assertFalse(result)

    def test_is_empty_AlwaysEmpty(self):
        cacher = DummyCacher()

        result = cacher.is_empty()

        self.assertTrue(result)

    def test_full_clear_AlwaysSuccess(self):
        cacher = DummyCacher()
        cacher.full_clear()


class TestInMemoryCacher(unittest.TestCase):

    def setUp(self) -> None:
        self.expected_result = 'OK'
        self.retriever_count = 0
        self.expiry_calculator = FakeExpiryCalculator(datetime.datetime(2000, 1, 1))
        self.cacher = InMemoryCacher(expiry_calculator=self.expiry_calculator)

    def retriever(self):
        self.retriever_count += 1
        return self.expected_result

    def test_expiry_DefaultExpiry(self):
        default_expiry = self.cacher.expiry

        self.assertEqual(default_expiry, self.cacher.default_expiry)

    def test_expiry_ExpiryChanges(self):
        new_expiry = datetime.timedelta(days=1)
        self.cacher.expiry = new_expiry

        self.assertEqual(self.cacher.expiry, new_expiry)

    def test_CahersHasNoCommonState(self):
        expiry_days = 1
        self.cacher.expiry = datetime.timedelta(days=expiry_days)

        # create second cacher
        another_cacher = InMemoryCacher(expiry_calculator=self.expiry_calculator)
        another_cacher.expiry = datetime.timedelta(days=expiry_days)

        # add items in first cache
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertFalse(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # second cache should stay empty
        self.assertFalse(another_cacher.has('', [], {}))
        self.assertTrue(another_cacher.is_empty())

    def test_ExpiresSuccess(self):
        now = self.expiry_calculator.now
        expiry_days = 1
        self.cacher.expiry = datetime.timedelta(days=expiry_days)

        # add items in cache
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertFalse(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # test that it's inside
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertTrue(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # two days later
        now += datetime.timedelta(days=expiry_days*2)
        self.expiry_calculator.now = now

        # cache should be cleaned
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertFalse(self.cacher.has('', [], {}))
        self.assertTrue(self.cacher.is_empty())

    def test_NotExpiresSuccess(self):
        now = self.expiry_calculator.now
        expiry_days = 4
        self.cacher.expiry = datetime.timedelta(days=expiry_days)

        # add items in cache
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertFalse(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # test that it's inside
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertTrue(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # two days later
        now += datetime.timedelta(days=expiry_days/2)
        self.expiry_calculator.now = now

        # items should persist
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

    def test_expiry_CleanedWhenChanged(self):
        now = self.expiry_calculator.now
        self.cacher.expiry = datetime.timedelta(days=1)

        # add items in cache
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertFalse(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # test that it's inside
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertTrue(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # another parameters - another cache item
        got_from_cache, result = self.cacher.retrieve('', [], {'a': 'a'}, self.retriever)

        self.assertFalse(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 2)  # incremented
        self.assertTrue(self.cacher.has('', [], {'a': 'a'}))
        self.assertFalse(self.cacher.is_empty())

        # one hour later
        minutes_spent = 60
        now += datetime.timedelta(minutes=minutes_spent)
        self.expiry_calculator.now = now

        # change expiry to half hour
        self.cacher.expiry = datetime.timedelta(minutes=minutes_spent/2)

        # cache should be cleaned
        self.assertEqual(self.retriever_count, 2)  # didn't changed
        self.assertFalse(self.cacher.has('', [], {}))
        self.assertTrue(self.cacher.is_empty())

    def test_drop_Success(self):
        self.cacher.expiry = datetime.timedelta(days=1)

        # add items in cache
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertFalse(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # test that it's inside
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertTrue(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # drop it
        was_in_cache = self.cacher.drop('', [], {})

        self.assertTrue(was_in_cache)
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertFalse(self.cacher.has('', [], {}))
        self.assertTrue(self.cacher.is_empty())

        # drop again
        was_in_cache = self.cacher.drop('', [], {})

        self.assertFalse(was_in_cache)
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertFalse(self.cacher.has('', [], {}))
        self.assertTrue(self.cacher.is_empty())

    def test_full_clear_Success(self):
        self.cacher.expiry = datetime.timedelta(days=1)

        # add items in cache
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertFalse(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # test that it's inside
        got_from_cache, result = self.cacher.retrieve('', [], {}, self.retriever)

        self.assertTrue(got_from_cache)
        self.assertEqual(result, self.expected_result)
        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertTrue(self.cacher.has('', [], {}))
        self.assertFalse(self.cacher.is_empty())

        # clear all
        self.cacher.full_clear()

        self.assertEqual(self.retriever_count, 1)  # didn't changed
        self.assertFalse(self.cacher.has('', [], {}))
        self.assertTrue(self.cacher.is_empty())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Base classes for data sources
"""

import abc
import decimal
import datetime
import typing

from ..communication.downloader import Downloader, DownloadStringResult

T = typing.TypeVar('T')  # pylint: disable=invalid-name


class SourceError(Exception):
    """ Base exception for all exceptions generated by sources. """

    def __init__(self, message: str):
        """ Initialize exception.

        :param message: Error message.
        """
        super().__init__(message)
        self.message = message


class SourceDownloadError(SourceError):
    """ Error happened while download data from source.
    """
    pass


class ParseError(SourceError):
    """ Error happened while parse downloaded data from source.
    """
    pass


class CheckApiActualityError(SourceError):
    """ Error happened while check API actuality of source.

        Uses as negative API actuality check result.
    """
    pass


class MaxPagesLimitExceeded(SourceError):
    """ Raise when limit of paged parameters intervals exceeded.
        Helps to prevent cycling.
    """
    def __init__(self, max_paged_parameters: int):
        """ Initialize exception.

        :param max_paged_parameters: Value of exceeded limit.
        """
        super().__init__(f"Limit of maximum paged parameters ({max_paged_parameters}) exceed.")


class InstrumentValuesHistoryEmpty(SourceError):
    """ Used by ``InstrumentValuesHistoryParser.parse`` to signal about empty history data
        if such data is acceptable.

        This is not signal about error but signal about acceptable empty sequence.
    """
    def __init__(self):
        """ Initialize exception.
        """
        super().__init__("Empty history data")


class InstrumentValue(typing.NamedTuple):
    """ Contains minimum (generic) data for value (e.g. price) representation
        of some financial instrument (stock, index etc.) on a moment in time.
    """
    value: decimal.Decimal
    moment: datetime.datetime


class InstrumentValueProvider(abc.ABC):
    """ Provides property for `InstrumentValue`
    """

    @abc.abstractmethod
    def get_instrument_value(self, tzinfo: typing.Optional[datetime.timezone]) -> InstrumentValue:
        """ Get generalized instrument history value.

        :param tzinfo: Expected time zone of moment inside instrument value.
        :return: `InstrumentValue` instance.
        """
        raise NotImplementedError


AnyInstrumentValueProvider = typing.TypeVar('AnyInstrumentValueProvider', bound=InstrumentValueProvider)


class InstrumentInfo(typing.NamedTuple):
    """ Contains minimum (generic) data for information (description)
        about some financial instrument (stock, index etc.).
    """
    code: str
    name: str


class InstrumentInfoProvider(abc.ABC):
    """ Provides property for `InstrumentInfo`
    """

    @property
    @abc.abstractmethod
    def instrument_info(self) -> InstrumentInfo:
        """ Get generalized instrument info.

        :return: Returns `InstrumentInfo`
        """
        raise NotImplementedError


AnyInstrumentInfoProvider = typing.TypeVar('AnyInstrumentInfoProvider', bound=InstrumentInfoProvider)


class InstrumentHistoryDownloadParameters(abc.ABC):
    """ Base class for instrument history download parameters
    """

    @abc.abstractmethod
    def clone_with_instrument_info_parameters(
            self,
            info_download_parameters: typing.Any,
            instrument_info: typing.Optional[InstrumentInfoProvider]) -> 'InstrumentHistoryDownloadParameters':
        """ Clones this history download parameters instance with replace some data
        with instrument info data.

        :param info_download_parameters: Optional instrument info download parameters for cloning.
        :param instrument_info: Optional instrument info for cloning.
        :return: Cloned history download parameters instance (self) with replacing some attributes from
                 `info_download_parameters` and `instrument_info`.
        """
        raise NotImplementedError


AnyInstrumentHistoryDownloadParameters = typing.TypeVar(
    'AnyInstrumentHistoryDownloadParameters',
    bound=InstrumentHistoryDownloadParameters)


class DynamicEnumTypeManager(abc.ABC):
    """ Base class for any dynamic enum types manager.

        Uses for static treatment of enum types, i.e. without web trips.
    """

    @abc.abstractmethod
    def is_dynamic_enum_type(self, cls: type) -> bool:
        """ Check whether ``cls`` is managed dynamic enum type.

        :param cls: Type (usually class) to check.
        :return: True if `cls` is managed dynamic enum type.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_all_managed_types(self) -> typing.Iterable[typing.Type]:
        """ Get all managed dynamic enum types.

        :return: All managed dynamic enum types.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_dynamic_enum_key(self, instance):
        """ Get dynamic enum key of instance.

        Returned key must be immutable and simply serializable,
        i.e. string, int, tuple, tuple of strings or ints, etc.

        :param instance: Instance of any managed dynamic enum type.
        :return: Dynamic enum key of `instance` or ``None`` if `instance` is not dynamic enum.
        """
        raise NotImplementedError


class DownloadParameterValuesStorage(DynamicEnumTypeManager, abc.ABC):
    """ Base class for storage of instrument download parameters.

        Uses for dynamically changing enum types, usually via web trips.
    """

    def reload(self) -> None:
        """ Force reload stored dynamic enum values, may be from external source.

        Method may access web.
        """
        pass

    def get_dynamic_enum_value_by_key(self, cls: type, key) -> typing.Any:
        """ Get value of dynamic enum type by its key.

        Key may be any value, uniquely identifying dynamic enum value of type 'cls'.
        Usually it's a some ID attribute.

        :param cls: Dynamic enum type (usually class) which value need to get.
        :param key: Key of required dynamic enum value.
        :return: Return value of dynamic enum type by its key.
                 `None` if 'cls' is not managed dynamic enum type.
        """
        return None

    def get_dynamic_enum_value_by_choice(self, cls: type, choice) -> typing.Any:
        """ Get value of dynamic enum type by its choice.

        Choice is stringified value of dynamic enum key.
        Usually uses in select inputs in HTML or any other API demanding string values
        (see `get_parameter_type_choices`).

        :param cls: Dynamic enum type (usually class) which value need to get.
        :param choice: Choice of required dynamic enum value.
        :return: Return value of dynamic enum type by its choice.
                 `None` if 'cls' is not managed dynamic enum type.
        """
        return None

    def get_all_parameter_values_for(self, cls: type) -> typing.Optional[typing.Iterable]:
        """ Get all values of dynamic enum type.

        :param cls: Dynamic enum type (usually class) which values need to get.
        :return: Return all values of dynamic enum type.
                 `None` if `cls` is not managed dynamic enum type.
        """
        return None

    def get_parameter_type_choices(self, cls: type) \
            -> typing.Optional[
                typing.List[typing.Tuple[typing.Any, typing.Union[str, typing.List[typing.Tuple[typing.Any, str]]]]]
            ]:
        """ If download parameters class contains attribute (maybe nested),
        which type allows only limited set of valid values,
        this function returns such set in form of sequence
        consisting itself of iterables of exactly two items
        (e.g. ``[(A, B), (A, B) ...]`` - a value ``A`` and a human-readable name ``B``)
        to use as choices for this attribute.

        You can also collect your available choices into named groups
        that can be used for organizational purposes::

            [('Audio', [('vinyl', 'Vinyl'), ('cd', 'CD'))],
             ('Video', [('vhs', 'VHS Tape'), ('dvd', 'DVD'))],
             ('unknown', 'Unknown')]

        The first element in each tuple is the name to apply to the group.
        The second element is an iterable of 2-tuples,
        with each 2-tuple containing a value and a human-readable name for an option.
        Grouped options may be combined with ungrouped options within a single list
        (such as the 'unknown' option in this example).

        Thus choices treats like select tag options (with optgroups) in HTML,
        or like the choices option for field in ``Django`` model API.

        :param cls: Dynamic enum type (usually class) which choices need to get.
        :return: Return all choices of dynamic enum type.
                 `None` if `cls` is unknown or doesn't need any special treatment.
        """
        return None


class InstrumentStringDataDownloader(abc.ABC):
    """ Base class for instrument data downloader.
    """

    @classmethod
    def adjust_date_from_to_holidays(cls, date_from: datetime.datetime) -> datetime.datetime:
        """ Utility function for adjusting `date from` value due to well known holidays.

        If `date from` locates into any holiday's interval, then shift it to nearest previous working day.
        The point is not to lose instrument history moment if its value calculates only at working days.

        Supports Christmas holidays and New Year's Eve.

        :param date_from: Date to adjust.
        :return: Adjusted `date from` value.
        """
        if date_from.month == 12 and date_from.day >= 25:
            # consider Christmas holidays: take last working day
            date_from = datetime.datetime(year=date_from.year, month=date_from.month, day=24, tzinfo=date_from.tzinfo)

        if date_from.month == 1 and date_from.day <= 10:
            # consider New Year's Eve: take last working day in previous year
            date_from = datetime.datetime(year=date_from.year - 1, month=12, day=24, tzinfo=date_from.tzinfo)

        return date_from

    def adjust_download_instrument_history_parameters(
            self,
            parameters: InstrumentHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Tuple[InstrumentHistoryDownloadParameters, datetime.datetime, datetime.datetime]:
        """ Takes parameters value and interval, analyzes it and returns similar parameters and interval,
        but adjusted to export interval dates
        (used in `InstrumentHistoryValuesExporter.export_instrument_history_values`)

        For instance, if 'moment_from' equals to 01-01-2000 (New Year holiday),
        then this method may adjust it to nearest working day, i.e. 24-12-1999,
        when instrument history value was set.

        Or, if downloader got interval of many years in parameters,
        but exporter need only couple of months,
        then this method will return similar parameters,
        but with smaller interval of dates in it, required to download requested data::

            adjust_download_instrument_history_parameters(
                {df='1967-01-01', dt='2020-12-31', ...},
                '2005-01-01',
                '2005-03-01'
                ) -> {df='2004-12-01', dt='2005-03-01', ...}, '2004-12-24', '2005-03-01'

        Notice that returned parameters interval may differ from requested by exporter interval
        (in this example ['2005-01-01', '2005-03-01'] vs ['2004-12-01', '2005-03-01'])
        It depends of downloader policy and dates themself.

        :param parameters: Source specific history download parameters to adjust.
        :param moment_from: Moment from to adjust.
        :param moment_to: Moment to to adjust.
        :return: Tuple of adjusted arguments.
        """
        if moment_from > moment_to:
            raise ValueError(f"Moment from ({moment_from}) is greater then moment to ({moment_to})")

        moment_from = self.adjust_date_from_to_holidays(moment_from)

        return parameters, moment_from, moment_to

    def paginate_download_instrument_history_parameters(
            self,
            parameters: InstrumentHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Iterable[typing.Tuple[InstrumentHistoryDownloadParameters, datetime.datetime, datetime.datetime]]:
        """ Takes parameters value and interval, analyzes it and returns sequence of similar parameters and interval,
        but paginated for downloading of most granular data in interval (daily usually).

        For instance, if downloader got interval of several years in parameters,
        but daily data can be downloaded only in one year interval,
        then this method will return sequence of similar parameters,
        but with one year interval of dates in them::

            paginate_download_instrument_history_parameters({df=2000, dt=2005, ...}) ->
                ({df=2000, dt=2000, ...}, {df=2001, dt=2001, ...}, {df=2002, dt=2002, ...},
                 {df=2003, dt=2003, ...}, {df=2004, dt=2004, ...}, {df=2005, dt=2005, ...})

        Default implementation does nothing and returns its arguments without modification.

        :param parameters: Source specific history download parameters to paginate.
        :param moment_from: Moment from to paginate.
        :param moment_to: Moment to to paginate.
        :return: Iterable of tuples of paginated arguments.
        """
        yield parameters, moment_from, moment_to

    @abc.abstractmethod
    def download_instrument_history_string(
            self,
            parameters: InstrumentHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime) -> DownloadStringResult:
        """ Downloads data of some instrument history as string and returns it.

        :param parameters: Source specific instrument history download parameters.
        :param moment_from: Download interval beginning.
        :param moment_to: Download interval ending.
        :return: Container with downloaded string.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def download_instruments_info_string(self, parameters) -> DownloadStringResult:
        """ Downloads data for set of instruments as string and returns it.

        :param parameters: Source specific instruments info download parameters.
        :return: Container with downloaded string.
        """
        raise NotImplementedError


class InstrumentValuesHistoryParser(abc.ABC):
    """ Parser for ``InstrumentValue`` history.
    """

    @abc.abstractmethod
    def parse(
            self,
            raw_text: str,
            tzinfo: typing.Optional[datetime.timezone]) -> typing.Iterable[InstrumentValueProvider]:
        """ Parses `raw_text` and returns `InstrumentValue` history.

        If `raw_text` is incorrect then raises ``ParseException``.
        If `raw_text` is correct, but contains unexpected or error-signalling content then raises ``DownloadError``.
        If parsed history data is empty (zero items) then raises ``InstrumentValuesHistoryEmpty``.

        :param raw_text: String to parse.
        :param tzinfo: Expected time zone of moment inside instrument value provider.
        :return: Iterable instrument values history.
        """
        raise NotImplementedError


class InstrumentInfoParser(abc.ABC):
    """ Parser for ``InstrumentInfo``.
    """

    @abc.abstractmethod
    def parse(self, raw_text: str) -> typing.Iterable[InstrumentInfoProvider]:
        """ Parses `raw_text` and returns ``InstrumentInfo`` sequence.

        If `raw_text` is incorrect then raises ``ParseException``.

        :param raw_text: String to parse.
        :return: Iterable instruments info.
        """
        raise NotImplementedError


class ApiActualityChecker(abc.ABC):
    """ Verifies actuality and accessibility of source API.
    """

    @abc.abstractmethod
    def check(self):
        """ Verifies actuality and accessibility of source API.

        Reads some data from source and verifies the result.
        If result is not expected or not relevant, then ``CheckApiActualityException`` raises.
        When success nothing happens (returns ``None``).
        """
        raise NotImplementedError


class InstrumentHistoryValuesExporter(abc.ABC):
    """ Base class for instrument history values exporter.
    """

    @abc.abstractmethod
    def export_instrument_history_values(
            self,
            parameters: InstrumentHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Iterable[InstrumentValueProvider]:
        """ Downloads data of some instrument history and returns it.

        :param parameters: Source specific instrument history download parameters.
        :param moment_from: Download interval beginning.
        :param moment_to: Download interval ending.
        :return: Iterable of instrument history values.
        """
        raise NotImplementedError


class InstrumentsInfoExporter(abc.ABC):
    """ Base class for instruments info exporter.
    """

    @abc.abstractmethod
    def export_instruments_info(self, parameters) -> typing.Iterator[InstrumentInfoProvider]:
        """ Downloads data for set of instruments info and returns it.

        :param parameters: Source specific instruments info download parameters.
        :return: Iterable of instruments info.
        """
        raise NotImplementedError


class DownloadParametersFactory(abc.ABC):
    """ Class for download parameters factories and generators.
    """

    @property
    @abc.abstractmethod
    def download_history_parameters_class(self) -> typing.Type[AnyInstrumentHistoryDownloadParameters]:
        """
        :return: Source specific class for download history parameters.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def download_history_parameters_factory(self) -> typing.Callable[..., AnyInstrumentHistoryDownloadParameters]:
        """
        :return: Factory for source specific download history parameters.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def download_info_parameters_class(self):
        """
        :return: Source specific class for download info parameters.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def download_info_parameters_factory(self) -> typing.Callable[..., typing.Any]:
        """
        :return: Factory for source specific download info parameters.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def generate_history_download_parameters_from(
            self,
            history_download_parameters: typing.Optional[AnyInstrumentHistoryDownloadParameters],
            info_download_parameters: typing.Any,
            instrument_info: typing.Optional[AnyInstrumentInfoProvider]) -> InstrumentHistoryDownloadParameters:
        """ Generate source specific history download parameters instance.

        :param history_download_parameters: Optional source specific history download parameters.
        :param info_download_parameters: Optional source specific info download parameters.
        :param instrument_info: Optional source specific instrument info.
        :return: New instance (cloned) of source specific history download parameters
                 with some attributes replaced from arguments.
        """
        raise NotImplementedError


class InstrumentExporterFactory(abc.ABC):
    """ Factory class for create instances of instrument data exporter.
    """

    @abc.abstractmethod
    def create_history_values_exporter(self, downloader: Downloader) -> InstrumentHistoryValuesExporter:
        """ Create new instance of source specific instrument history values exporter.

        :param downloader: Downloader used for get instrument data.
        :return: New instance of source specific instrument history values exporter.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_info_exporter(self, downloader: Downloader) -> InstrumentsInfoExporter:
        """ Create new instance of source specific instruments info exporter.

        :param downloader: Downloader used for get instrument data.
        :return: New instance of source specific instruments info exporter.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_download_parameter_values_storage(self, downloader: Downloader) -> DownloadParameterValuesStorage:
        """ Create new instance of source specific instrument download parameters storage.

        :param downloader: Downloader used for get storage data.
        :return: New instance of source specific instrument download parameters storage.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_api_actuality_checker(self, downloader: Downloader) -> ApiActualityChecker:
        """ Create new instance of source specific API actuality checker.

        :param downloader: Downloader used for get API data.
        :return: New instance of source specific API actuality checker.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dynamic_enum_type_manager(self) -> DynamicEnumTypeManager:
        """ Get source specific dynamic enum type manager.

        :return: Instance of source specific dynamic enum type manager.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def download_parameters_factory(self) -> DownloadParametersFactory:
        """ Get source specific download parameters factory.

        :return: Source specific download parameters factory.
        """
        raise NotImplementedError


class InstrumentExporterRegistry(typing.NamedTuple):
    """ Container for information about registered ``InstrumentHistoryValuesExporterFactory``.
    """
    factory: InstrumentExporterFactory
    name: str
    provider_site: typing.Optional[str] = None
    api_url: typing.Optional[str] = None



""" Sources from Solactive.
"""
from . import v2018
from ..generic import InstrumentExporterRegistry, register_instrument_history_values_exporter

__all__ = ['v2018']

register_instrument_history_values_exporter(
    InstrumentExporterRegistry(
        name=v2018.exporters.SolactiveExporterFactory.name,
        provider_site=v2018.exporters.SolactiveExporterFactory.provider_site,
        factory=v2018.exporters.SolactiveExporterFactory()
    ))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import unittest

from sane_finances.sources.base import CheckApiActualityError, InstrumentExporterFactory, ParseError
from sane_finances.sources.cbr.v2016.exporters import CbrStringDataDownloader, CbrApiActualityChecker, \
    CbrCurrencyRatesExporterFactory, CbrCurrencyDownloadParameterValuesStorage
from sane_finances.sources.cbr.v2016.meta import (
    RateFrequencies, CurrencyInfo, CurrencyRateValue,
    CbrCurrenciesInfoDownloadParameters, CbrCurrencyHistoryDownloadParameters)
from .common import CommonTestCases
from .fakes import (
    FakeDownloader, FakeCbrCurrencyInfoParser,
    FakeCbrCurrencyHistoryXmlParser, FakeCbrStringDataDownloader)


class TestMsciStringDataDownloader(unittest.TestCase):

    def setUp(self):
        # fake downloaders
        self.fake_data = 'data'

        downloader = FakeDownloader(self.fake_data)

        self.string_data_downloader = CbrStringDataDownloader(downloader)

    def test_download_currency_history_string(self):
        currency_id = 'USD'
        date_from = datetime.date(2010, 1, 1)
        date_to = date_from + datetime.timedelta(days=365*10)

        result = self.string_data_downloader.download_currency_history_string(
            currency_id,
            date_from,
            date_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_instrument_history_string(self):
        currency_id = 'USD'
        moment_from = datetime.datetime(2010, 1, 1)  # no hours
        moment_to = moment_from + datetime.timedelta(days=365*10)
        parameters = CbrCurrencyHistoryDownloadParameters.safe_create(currency_id=currency_id)

        result = self.string_data_downloader.download_instrument_history_string(parameters, moment_from, moment_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

        moment_from = datetime.datetime(2010, 1, 1, 12)  # has hours
        moment_to = moment_from + datetime.timedelta(days=365*10)
        parameters = CbrCurrencyHistoryDownloadParameters.safe_create(currency_id=currency_id)

        result = self.string_data_downloader.download_instrument_history_string(parameters, moment_from, moment_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_currencies_info_string(self):
        rate_frequency = RateFrequencies.DAILY

        result = self.string_data_downloader.download_currencies_info_string(rate_frequency)

        self.assertSequenceEqual(result.downloaded_string, self.fake_data)

    def test_download_instruments_info_string(self):
        rate_frequency = RateFrequencies.DAILY
        parameters = CbrCurrenciesInfoDownloadParameters.safe_create(rate_frequency=rate_frequency)

        result = self.string_data_downloader.download_instruments_info_string(parameters)

        self.assertSequenceEqual(result.downloaded_string, self.fake_data)


class TestCbrAPIActualityChecker(unittest.TestCase):

    def setUp(self):
        # fakes
        self.fake_info_str_data = 'info'
        self.fake_history_str_data = 'history'

        self.success_history_parsed_item = \
            CurrencyRateValue(
                date=CbrApiActualityChecker._expectedFirstDate,
                value=CbrApiActualityChecker._expectedFirstValue,
                nominal=CbrApiActualityChecker._expectedFirstNominal,
                currency_id=CbrApiActualityChecker._currencyToCheck
            )
        self.success_info_parsed_item = CurrencyInfo(
            currency_id='FAKE_ID',
            name='FAKE_NAME',
            eng_name='FAKE_ENG_NAME',
            nominal=1,
            parent_code='FAKE_PARENT_CODE')

        self.fake_string_data_downloader = FakeCbrStringDataDownloader(
            self.fake_info_str_data,
            self.fake_history_str_data)

    def test_check_Success(self):
        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])
        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        checker.check()

    def test_check_raisesWhenInfoParseError(self):
        # corrupt data
        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])
        info_parser.parse_exception = ParseError('Error')
        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenInfoUnknownError(self):
        # corrupt data
        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])
        info_parser.parse_exception = Exception('Error')
        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenNoInfo(self):
        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeCbrCurrencyInfoParser([])  # No data

        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenHistoryParseError(self):
        # corrupt data
        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        history_parser.parse_exception = ParseError('Error')
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])
        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenHistoryUnknownError(self):
        # corrupt data
        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        history_parser.parse_exception = Exception('Error')
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])
        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.fake_string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.fake_string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenWrongCurrencyId(self):
        # corrupt data
        self.success_history_parsed_item.currency_id = 'WRONG'

        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])

        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongNominal(self):
        # corrupt data
        self.success_history_parsed_item.nominal = -1

        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])

        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongFirstDate(self):
        # corrupt data
        self.success_history_parsed_item.date = datetime.date.max

        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])

        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongFirstValue(self):
        # corrupt data
        self.success_history_parsed_item.value = decimal.Decimal(-1)

        history_parser = FakeCbrCurrencyHistoryXmlParser([self.success_history_parsed_item])
        info_parser = FakeCbrCurrencyInfoParser([self.success_info_parsed_item])

        checker = CbrApiActualityChecker(self.fake_string_data_downloader, history_parser, info_parser)

        with self.assertRaises(CheckApiActualityError):
            checker.check()


class TestCbrCurrencyDownloadParameterValuesStorage(unittest.TestCase):

    def setUp(self):
        self.storage = CbrCurrencyDownloadParameterValuesStorage()

    def test_is_dynamic_enum_type_AlwaysFalse(self):
        self.assertFalse(self.storage.is_dynamic_enum_type(RateFrequencies))
        # noinspection PyTypeChecker
        self.assertFalse(self.storage.is_dynamic_enum_type(None))

    def test_get_dynamic_enum_key_AlwaysNone(self):
        self.assertIsNone(self.storage.get_dynamic_enum_key(RateFrequencies.DAILY))

    def test_get_all_managed_types_AlwaysEmpty(self):
        self.assertSequenceEqual(list(self.storage.get_all_managed_types()), [])

    def test_get_parameter_type_choices_Success(self):
        all_types = self.storage._special_handlers.keys()
        for dynamic_enum_type in all_types:
            choices = self.storage.get_parameter_type_choices(dynamic_enum_type)

            self.assertIsNotNone(choices)

        # noinspection PyTypeChecker
        choices = self.storage.get_parameter_type_choices(None)

        self.assertIsNone(choices)

        choices = self.storage.get_parameter_type_choices(object)

        self.assertIsNone(choices)


class TestCbrCurrencyRatesExporterFactory(CommonTestCases.CommonIndexExporterFactoryTests):

    def get_exporter_factory(self) -> InstrumentExporterFactory:
        return CbrCurrencyRatesExporterFactory()

    def is_dynamic_enum_type_manager_singleton(self):
        return True

    def is_download_parameters_factory_singleton(self):
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import decimal
import unittest

from sane_finances.sources.base import (
    InstrumentValue, InstrumentInfo, DownloadParametersFactory, InstrumentHistoryDownloadParameters)
from sane_finances.sources.solactive.v2018.meta import (
    IndexValue, IndexInfo, SolactiveIndexesInfoDownloadParameters, SolactiveIndexHistoryDownloadParameters,
    SolactiveDownloadParametersFactory)
from .common import CommonTestCases


class TestIndexValue(unittest.TestCase):

    def test_instrument_value_Success(self):
        index_value = IndexValue(
            index_id='ID',
            moment=datetime.datetime(2000, 12, 31, 12, tzinfo=datetime.timezone.utc),
            value=decimal.Decimal(42))
        expected_instrument_value = InstrumentValue(
            value=index_value.value,
            moment=index_value.moment)

        instrument_value = index_value.get_instrument_value(datetime.timezone.utc)

        self.assertEqual(expected_instrument_value, instrument_value)

    def test_raiseWrongMoment(self):
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            _ = IndexValue(
                index_id='ID',
                moment=None,
                value=decimal.Decimal(42))


class TestIndexInfo(unittest.TestCase):

    def test_instrument_value_Success(self):
        index_info = IndexInfo(
            isin='ISIN',
            name='NAME')
        expected_instrument_info = InstrumentInfo(code=index_info.isin, name=index_info.name)

        instrument_info = index_info.instrument_info

        self.assertEqual(expected_instrument_info, instrument_info)


class TestSolactiveIndexesInfoDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = SolactiveIndexesInfoDownloadParameters.safe_create()


class TestSolactiveIndexHistoryDownloadParameters(unittest.TestCase):

    def test_safe_create_Success(self):
        _ = SolactiveIndexHistoryDownloadParameters.safe_create(isin='ISIN')

    def test_clone_with_instrument_info_parameters_Success(self):
        params = SolactiveIndexHistoryDownloadParameters.safe_create(isin='ISIN')
        expected_result = params

        result = params.clone_with_instrument_info_parameters(None, None)

        self.assertEqual(expected_result, result)


class TestSolactiveDownloadParametersFactory(CommonTestCases.CommonDownloadParametersFactoryTests):

    def get_download_parameters_factory(self) -> DownloadParametersFactory:
        return SolactiveDownloadParametersFactory()

    def get_generate_history_download_parameters_from_expected_result_with_none(self) \
            -> InstrumentHistoryDownloadParameters:
        # noinspection PyTypeChecker
        expected_result = SolactiveIndexHistoryDownloadParameters(isin=None)
        return expected_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Cachers for downloaders
"""

import abc
import logging
import datetime
import typing

logging.getLogger().addHandler(logging.NullHandler())


class ExpiryCalculator:
    """ Uses for calculations of expiry and revive moments.
    """

    def is_expired(self, expiry_moment: datetime.datetime) -> bool:
        """ Verify `expiry_moment` for expiration.

        :param expiry_moment: Moment to verify.
        :return: ``True`` if 'expiry_moment' was in past (i.e. expired).
        """
        return datetime.datetime.now(tz=datetime.timezone.utc) > expiry_moment

    def get_expiry_moment(self, delta: datetime.timedelta, start_from: datetime.datetime = None) -> datetime.datetime:
        """ Calculate and return moment shifted on `delta` time from `start_from` moment.
            If `start_from` is ``None`` then moment shifts from current moment (from now).

        :param delta: Timedelta for calculation.
        :param start_from: Start moment for calculation.
        :return: Calculated expiry moment relatively to `start_from` on `delta` time.
        """
        if start_from is None:
            start_from = datetime.datetime.now(tz=datetime.timezone.utc)
        return start_from + delta

    def get_revive_moment(self) -> datetime.datetime:
        """ Usually return current moment (now).

        :return: Moment of this method invoking.
        """
        return datetime.datetime.now(tz=datetime.timezone.utc)


class BaseCacher(abc.ABC):
    """ Base class (interface) for all cachers.
    """

    @abc.abstractmethod
    def retrieve(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str],
            reviver: typing.Callable[[], str]) -> typing.Tuple[bool, str]:
        """ Try to find previously downloaded string by arguments inside the internal storage.
        If not found, then call `reviver` and store it result.

        :param url: URL of request.
        :param parameters: Parameters of request.
        :param headers: Headers of request.
        :param reviver: Function that returns downloaded string if such not found in cache yet.
        :return: Pair: (got_from_cache, result)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def drop(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]) -> bool:
        """ Drop items from internal storage by parameters.

        :param url: URL of request.
        :param parameters: Parameters of request.
        :param headers: Headers of request.
        :return: ``True`` if any item was dropped. ``False`` if no items was in storage by that parameters.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def has(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]) -> bool:
        """ Verify that cache contains items with such parameters.

        :param url: URL of request.
        :param parameters: Parameters of request.
        :param headers: Headers of request.
        :return: ``True`` if there are any item is in storage by that parameters.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_empty(self) -> bool:
        """ Verify that cache is empty.

        :return: ``True`` if cache is empty
        """
        raise NotImplementedError

    @abc.abstractmethod
    def full_clear(self):
        """ Full clear of internal storage.
        """
        raise NotImplementedError


class ExpirableCacher(BaseCacher, abc.ABC):
    """ Base class for cachers that can expire their stored items.
    """
    @property
    @abc.abstractmethod
    def expiry(self) -> datetime.timedelta:
        """ Current timedelta value of expiry.
        """
        raise NotImplementedError

    @expiry.setter
    @abc.abstractmethod
    def expiry(self, delta: datetime.timedelta):
        """ Set new value of timedelta for expiry.

        Usually calls `clean` for revisiting stored items due to new expiry value.

        :param delta: New value.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def clean(self):
        """ Revisit all cached items, drop expired, update expiry moment if needed.
        """
        raise NotImplementedError


class DummyCacher(BaseCacher):
    """ Cacher without cache.
    """

    def retrieve(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str],
            reviver: typing.Callable[[], str]) -> typing.Tuple[bool, str]:
        return False, reviver()

    def drop(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]) -> bool:
        return False

    def has(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]) -> bool:
        return False

    def is_empty(self) -> bool:
        return True

    def full_clear(self):
        pass


class _InMemoryCacheKey(typing.NamedTuple):
    url: str
    parameters: typing.Tuple[typing.Tuple[str, str]]
    headers: typing.Tuple[typing.Tuple[str, str]]


class _InMemoryCacheItem(typing.NamedTuple):
    key: _InMemoryCacheKey
    result: str
    revive_moment: datetime.datetime
    expiry_moment: datetime.datetime


class InMemoryCacher(ExpirableCacher):
    """ In-memory cacher based on dictionary.
    """
    default_expiry: datetime.timedelta = datetime.timedelta()  # zero delta

    _expiry: datetime.timedelta
    _storage: typing.Dict[_InMemoryCacheKey, _InMemoryCacheItem]

    def __init__(self,
                 initial_expiry: datetime.timedelta = None,
                 expiry_calculator: ExpiryCalculator = None):
        """ Initialize cacher.

        :param initial_expiry: Initial value of expiry.
        :param expiry_calculator: Expiry calculator used for such calculations.
        """
        self.expiry_calculator = ExpiryCalculator() if expiry_calculator is None else expiry_calculator

        self._expiry = self.default_expiry if initial_expiry is None else initial_expiry
        self._storage = {}

    @property
    def expiry(self) -> datetime.timedelta:
        return self._expiry

    @expiry.setter
    def expiry(self, delta: datetime.timedelta):
        self._expiry = delta
        self.clean()

    @staticmethod
    def _build_key(
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]):
        key = _InMemoryCacheKey(
            url=str(url),
            parameters=tuple((str(param_name), str(param_value)) for param_name, param_value in parameters),
            headers=tuple((str(header_name), str(header_value)) for header_name, header_value in headers.items())
        )
        return key

    def clean(self):
        items = tuple(self._storage.values())
        for item in items:
            new_expiry_moment = self.expiry_calculator.get_expiry_moment(
                self.expiry,
                start_from=item.revive_moment)

            if (self.expiry_calculator.is_expired(item.expiry_moment)
                    or self.expiry_calculator.is_expired(new_expiry_moment)):
                del self._storage[item.key]

            else:
                self._storage[item.key] = item._replace(expiry_moment=new_expiry_moment)

    def retrieve(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str],
            reviver: typing.Callable[[], str]) -> typing.Tuple[bool, str]:

        self.clean()

        key = self._build_key(url, parameters, headers)
        got_from_cache, cache_item = True, self._storage.get(key, None)
        if cache_item is None or self.expiry_calculator.is_expired(cache_item.expiry_moment):
            got_from_cache, result = False, reviver()
            revive_moment = self.expiry_calculator.get_revive_moment()
            expiry_moment = self.expiry_calculator.get_expiry_moment(self.expiry)
            self._storage[key] = cache_item = _InMemoryCacheItem(
                key=key,
                result=result,
                revive_moment=revive_moment,
                expiry_moment=expiry_moment)

        return got_from_cache, cache_item.result

    def drop(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]) -> bool:

        self.clean()

        key = self._build_key(url, parameters, headers)
        if key in self._storage:
            del self._storage[key]
            return True

        return False

    def has(
            self,
            url: str,
            parameters: typing.List[typing.Tuple[str, str]],
            headers: typing.Dict[str, str]) -> bool:

        self.clean()

        key = self._build_key(url, parameters, headers)
        return key in self._storage

    def is_empty(self) -> bool:
        self.clean()

        return len(self._storage) == 0

    def full_clear(self):
        self._storage.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Utilities for parse data from moex.com.
"""
import decimal
import json
import logging
import typing
import datetime

from .meta import SecurityValue, SecurityInfo, GlobalIndexData, TradeEngine, Market, Board
from ...base import (
    InstrumentValuesHistoryParser, InstrumentInfoParser, InstrumentValuesHistoryEmpty, DownloadParameterValuesStorage,
    ParseError)

logging.getLogger().addHandler(logging.NullHandler())


def _parse_block(
        block_name: str,
        raw_data: typing.Dict,
        attrs_mapping: typing.Iterable[typing.Tuple[str, str, bool, typing.Any]],
        raise_when_data_block_is_empty=False) -> typing.Iterable[typing.Dict]:

    block = raw_data.get(block_name, None)
    if block is None:
        raise ParseError(f"Wrong JSON format. '{block_name}' block not found.")

    columns_block = block.get('columns', None)
    if columns_block is None:
        raise ParseError(f"Wrong JSON format. '{block_name}.columns' block not found.")
    if not isinstance(columns_block, list):
        raise ParseError(f"Wrong JSON format. '{block_name}.columns' block is not list.")

    data_block = block.get('data', None)
    if data_block is None:
        raise ParseError(f"Wrong JSON format. '{block_name}.data' block not found.")
    if not isinstance(data_block, list):
        raise ParseError(f"Wrong JSON format. '{block_name}.data' block is not list.")

    if not data_block and raise_when_data_block_is_empty:
        raise InstrumentValuesHistoryEmpty()

    data_mapping = []
    for attr_name, column_name, is_required, converter in attrs_mapping:
        try:
            column_index = columns_block.index(column_name)
        except ValueError as ex:
            if is_required:
                raise ParseError(f"Wrong JSON format. "
                                 f"Column {column_name!r} not found in '{block_name}.columns' block.") from ex
        else:
            data_mapping.append((attr_name, converter, column_index))

    for data_item in data_block:
        factory_kwargs = {}
        for attr_name, converter, column_index in data_mapping:
            try:
                attr_value = data_item[column_index]
            except IndexError as ex:
                raise ParseError(f"Wrong JSON format. "
                                 f"Column with index {column_index!r} not found in '{block_name}.data' block: "
                                 f"{data_item}") from ex

            if converter is not None:
                attr_value = converter(attr_value)

            factory_kwargs[attr_name] = attr_value

        yield factory_kwargs


class MoexHistoryJsonParser(InstrumentValuesHistoryParser):
    """ Parser for history data of JSON string.

    Accept JSON like::

            {
            "history": {
                "columns": ["BOARDID", "TRADEDATE", "SECID", "CLOSE", "LEGALCLOSEPRICE", "FACEVALUE"],
                "data": [
                    ["TQBR", "2014-06-25", "ABRD", 134.12, 134.12, 1000],
                    ...
                ]
            }}
    """

    trade_date_format = '%Y-%m-%d'

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self._attrs_mapping = (
            ('trade_date', 'TRADEDATE', True, self._convert_trade_date_to_date),
            ('legal_close_price', 'LEGALCLOSEPRICE', False, self._convert_price_to_decimal),
            ('close', 'CLOSE', False, self._convert_price_to_decimal),
            ('face_value', 'FACEVALUE', False, self._convert_price_to_decimal)
        )

    def parse(  # pylint: disable=arguments-renamed
            self,
            raw_json_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[SecurityValue]:

        try:
            raw_data = json.loads(raw_json_text)
        except json.decoder.JSONDecodeError as ex:
            raise ParseError(ex.msg) from ex

        if not isinstance(raw_data, dict):
            # can be Inf etc., but we accept only {}
            raise ParseError("Wrong JSON format. Top level is not dictionary.")

        for factory_kwargs in _parse_block(
                'history',
                raw_data,
                self._attrs_mapping,
                raise_when_data_block_is_empty=True):

            if 'legal_close_price' not in factory_kwargs and 'close' not in factory_kwargs:
                raise ParseError("Wrong JSON format. Neither 'LEGALCLOSEPRICE' nor 'CLOSE' column was found.")

            close = factory_kwargs.get('close', None)
            legal_close_price = factory_kwargs.get('legal_close_price', None)

            if legal_close_price is not None or close is not None:
                close = legal_close_price if legal_close_price is not None else close

                if 'face_value' in factory_kwargs:
                    face_value: decimal.Decimal = factory_kwargs['face_value']
                    if face_value is not None:
                        close *= face_value/100

                    del factory_kwargs['face_value']

                if 'legal_close_price' in factory_kwargs:
                    del factory_kwargs['legal_close_price']

                factory_kwargs['close'] = close
                yield SecurityValue(**factory_kwargs)

    def _convert_trade_date_to_date(self, trade_date: str):
        try:
            trade_date = datetime.datetime.strptime(trade_date, self.trade_date_format)
        except (ValueError, TypeError) as ex:
            raise ParseError(f"Wrong JSON format. "
                             f"Can't convert {trade_date!r} to date.") from ex

        trade_date = trade_date.date()
        return trade_date

    @staticmethod
    def _convert_price_to_decimal(price):
        if price is None:
            return None

        if isinstance(price, float):
            # hack to adjust floating point digits
            price = repr(price)

        try:
            decimal_price = decimal.Decimal(price)
        except (ValueError, TypeError, decimal.DecimalException) as ex:
            raise ParseError(f"Wrong JSON format. "
                             f"Can't convert {price!r} to decimal.") from ex

        return decimal_price


class MoexSecurityInfoJsonParser(InstrumentInfoParser):
    """ Parser for security info list from JSON.

    Accept JSON like::

            {
                "securities": {
                    ...
                    "columns": ["SECID", "BOARDID", "SHORTNAME", ...],
                    "data": [
                                ["AKMD", "TQTD", "ETF AKMD", ...],
                                ...
                            ]
                    ...
                },
                ...
            }
    """

    def __init__(self, parameter_values_storage: DownloadParameterValuesStorage):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.parameter_values_storage = parameter_values_storage

        self._attrs_mapping = (
            ('sec_id', 'SECID', True, None),
            ('board', 'BOARDID', True, self._get_board),
            ('short_name', 'SHORTNAME', True, None),
            ('lot_size', 'LOTSIZE', False, None),
            ('sec_name', 'SECNAME', False, None),
            ('sec_name', 'NAME', False, None),
            ('isin', 'ISIN', False, None),
            ('lat_name', 'LATNAME', False, None),
            ('reg_number', 'REGNUMBER', False, None)
        )
        self._boards = {}

    def parse(self, raw_json_text: str) -> typing.Iterable[SecurityInfo]:  # pylint: disable=arguments-renamed
        self._boards = {
            board.boardid: board
            for board
            in self.parameter_values_storage.get_all_parameter_values_for(Board)}

        try:
            raw_data = json.loads(raw_json_text)
        except json.decoder.JSONDecodeError as ex:
            raise ParseError(ex.msg) from ex

        if not isinstance(raw_data, dict):
            # can be Inf etc., but we accept only {}
            raise ParseError("Wrong JSON format. Top level is not dictionary.")

        return (SecurityInfo(**factory_kwargs)
                for factory_kwargs
                in _parse_block('securities', raw_data, self._attrs_mapping))

    def _get_board(self, boardid: str):
        if boardid not in self._boards:
            raise ParseError(f"Board {boardid} not found.")

        return self._boards[boardid]


class MoexGlobalIndexJsonParser:
    """ Parser for global index data from JSON.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self._engines_attrs_mapping = (
            ('identity', 'id', True, None),
            ('name', 'name', True, None),
            ('title', 'title', True, None)
        )
        self._markets_attrs_mapping = (
            ('identity', 'id', True, None),
            ('trade_engine', 'trade_engine_id', True, self._get_engine_by_id),
            ('name', 'market_name', True, None),
            ('title', 'market_title', True, None),
            ('marketplace', 'marketplace', True, None)
        )
        self._board_groups_attrs_mapping = (
            ('identity', 'id', True, None),
            ('trade_engine', 'trade_engine_id', True, self._get_engine_by_id),
            ('market', 'market_id', True, self._get_market_by_id),
            ('name', 'name', True, None),
            ('title', 'title', True, None),
            ('is_default', 'is_default', True, bool),
            ('is_traded', 'is_traded', True, bool)
        )
        self._boards_attrs_mapping = (
            ('identity', 'id', True, None),
            ('trade_engine', 'engine_id', True, self._get_engine_by_id),
            ('market', 'market_id', True, self._get_market_by_id),
            ('boardid', 'boardid', True, None),
            ('title', 'board_title', True, None),
            ('is_traded', 'is_traded', True, bool),
            ('has_candles', 'has_candles', True, bool),
            ('is_primary', 'is_primary', True, bool)
        )
        self._engines = {}
        self._markets = {}

    def parse(self, raw_json_text: str) -> GlobalIndexData:
        """ Parse global index data from JSON and return it.

        :param raw_json_text: JSON string with global index data.
        :return: ``GlobalIndexData`` instance.
        """
        try:
            raw_data = json.loads(raw_json_text)
        except json.decoder.JSONDecodeError as ex:
            raise ParseError(ex.msg) from ex

        if not isinstance(raw_data, dict):
            # can be Inf etc., but we accept only {}
            raise ParseError("Wrong JSON format. Top level is not dictionary.")

        engines = tuple(TradeEngine(**factory_kwargs)
                        for factory_kwargs
                        in _parse_block('engines', raw_data, self._engines_attrs_mapping))
        self._engines = {trade_engine.identity: trade_engine for trade_engine in engines}

        markets = tuple(Market(**factory_kwargs)
                        for factory_kwargs
                        in _parse_block('markets', raw_data, self._markets_attrs_mapping))
        self._markets = {market.identity: market for market in markets}

        boards = tuple(Board(**factory_kwargs)
                       for factory_kwargs
                       in _parse_block('boards', raw_data, self._boards_attrs_mapping))

        return GlobalIndexData(trade_engines=engines, markets=markets, boards=boards)

    def _get_engine_by_id(self, trade_engine_id: int):
        if trade_engine_id not in self._engines:
            raise ParseError(f"Trade engine with id = {trade_engine_id} not found.")

        return self._engines[trade_engine_id]

    def _get_market_by_id(self, market_id: int):
        if market_id not in self._markets:
            raise ParseError(f"Market with id = {market_id} not found.")

        return self._markets[market_id]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Tools for download (export) index data from app2.msci.com
"""
import logging
import typing
import datetime
import decimal
import inspect

from .meta import (
    Market, Currency, IndexLevel, Frequency, Style, Size, Scopes,
    IndexSuite, IndexSuiteGroup, IndexPanelData,
    MsciIndexHistoryDownloadParameters, MsciIndexesInfoDownloadParameters,
    MsciDownloadParametersFactory)
from .parsers import MsciHistoryJsonParser, MsciIndexInfoParser, MsciIndexPanelDataJsonParser
from ...base import (
    ApiActualityChecker, InstrumentStringDataDownloader, ParseError,
    InstrumentExporterFactory, InstrumentHistoryValuesExporter, InstrumentsInfoExporter, DynamicEnumTypeManager,
    DownloadParameterValuesStorage, CheckApiActualityError)
from ...generic import GenericInstrumentHistoryValuesExporter, GenericInstrumentsInfoExporter
from ....communication.downloader import Downloader, DownloadStringResult

logging.getLogger().addHandler(logging.NullHandler())


class MsciDynamicEnumTypeManager(DynamicEnumTypeManager):
    """ MSCI dynamic enum types manager.
    """

    def __init__(self):
        self._managed_types: typing.Dict[typing.Type, typing.Tuple] = {
            Market: ('markets', lambda it: it.identity),
            Currency: ('currencies', lambda it: it.identity),
            IndexLevel: ('index_levels', lambda it: it.identity),
            Frequency: ('frequencies', lambda it: it.identity),
            IndexSuiteGroup: ('index_suite_groups', lambda it: it.name),
            IndexSuite: ('index_suites', lambda it: it.identity),
            Size: ('sizes', lambda it: it.identity),
            Style: ('styles', lambda it: it.identity),
        }

    def is_dynamic_enum_type(self, cls: type) -> bool:
        if not inspect.isclass(cls):
            return False

        return cls in self._managed_types

    def get_all_managed_types(self) -> typing.Iterable[typing.Type]:
        return self._managed_types.keys()

    def get_dynamic_enum_key(self, instance):
        for managed_type, (_, key_getter, *_) in self._managed_types.items():
            if isinstance(instance, managed_type):
                return key_getter(instance)

        return None


class MsciIndexDownloadParameterValuesStorage(MsciDynamicEnumTypeManager, DownloadParameterValuesStorage):
    """ Storage of indexes download parameters.
    """
    index_panel_data_url = \
        'https://app2.msci.com/products/index-data-search/resources/index-data-search/js/data/index-panel-data.json'

    def __init__(
            self,
            downloader: Downloader,
            index_panel_data_json_parser: MsciIndexPanelDataJsonParser):
        super().__init__()
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.downloader = downloader
        self.index_panel_data_json_parser = index_panel_data_json_parser

        # headers for HTTP
        self.headers: typing.Dict[str, str] = {}

        self._extended_managed_types: typing.Dict[typing.Type, typing.Tuple] = {
            Market: (self._get_markets_choices,),
            Currency: (self._get_currencies_choices,),
            IndexLevel: (self._get_index_levels_choices,),
            Frequency: (self._get_frequencies_choices,),
            IndexSuiteGroup: (self._get_index_suite_groups_choices,),
            IndexSuite: (self._get_index_suites_choices,),
            Size: (self._get_sizes_choices,),
            Style: (self._get_styles_choices,)
        }
        assert set(self._extended_managed_types.keys()) == set(self._managed_types.keys()), \
            ("MoexDynamicEnumTypeManager and MoexDownloadParameterValuesStorage "
             "has different managed types.")

        self.index_panel_data: typing.Optional[IndexPanelData] = None

    def reload(self) -> None:
        self.downloader.headers = self.headers
        self.downloader.parameters = []
        json_string_result = self.downloader.download_string(self.index_panel_data_url)
        try:
            index_panel_data = self.index_panel_data_json_parser.parse(json_string_result.downloaded_string)

        except Exception:
            json_string_result.set_correctness(False)
            raise

        json_string_result.set_correctness(True)
        self.index_panel_data = index_panel_data

    def _ensure_loaded(self):
        if self.index_panel_data is None:
            self.reload()

    @property
    def daily_frequency(self) -> Frequency:
        """ Value of daily frequency.
        """
        self._ensure_loaded()

        return self.index_panel_data.daily_frequency

    @property
    def monthly_frequency(self) -> Frequency:
        """ Value of monthly frequency.
        """
        self._ensure_loaded()

        return self.index_panel_data.monthly_frequency

    def _get_dynamic_enum_value_by_key(self, cls: type, key):
        """ """
        index_panel_data_attr_name, key_getter, *_ = self._managed_types[cls]
        enum_values = getattr(self.index_panel_data, index_panel_data_attr_name)

        for enum_value in enum_values:
            if key == key_getter(enum_value):
                return enum_value

        return None

    def get_dynamic_enum_value_by_key(self, cls: type, key) -> typing.Any:
        if not self.is_dynamic_enum_type(cls):
            return None

        self._ensure_loaded()

        return self._get_dynamic_enum_value_by_key(cls, key)

    def get_dynamic_enum_value_by_choice(self, cls: type, choice) -> typing.Any:
        if not self.is_dynamic_enum_type(cls):
            return None

        self._ensure_loaded()

        return self._get_dynamic_enum_value_by_key(cls, str(choice))

    def get_all_parameter_values_for(self, cls: type) -> typing.Optional[typing.Iterable]:
        if not self.is_dynamic_enum_type(cls):
            return None

        self._ensure_loaded()

        index_panel_data_attr_name, *_ = self._managed_types[cls]
        return getattr(self.index_panel_data, index_panel_data_attr_name)

    def get_parameter_type_choices(self, cls: type) \
            -> typing.Optional[
                typing.List[typing.Tuple[typing.Any, typing.Union[str, typing.List[typing.Tuple[typing.Any, str]]]]]
            ]:
        if not self.is_dynamic_enum_type(cls):
            return None

        self._ensure_loaded()

        choices_getter, *_ = self._extended_managed_types[cls]
        return choices_getter()

    def _get_markets_choices(self):
        markets_without_scope = [(market.identity, market.name)
                                 for market
                                 in self.index_panel_data.markets
                                 if market.scope is None]
        scopes = set(market.scope for market in self.index_panel_data.markets if market.scope is not None)
        scoped_markets = [(f"Available for {scope.description!r} scope:",
                           [(market.identity, market.name)
                            for market
                            in self.index_panel_data.markets
                            if market.scope == scope])
                          for scope in scopes]

        # noinspection PyTypeChecker
        return markets_without_scope + scoped_markets

    def _get_currencies_choices(self):
        return [(currency.identity, currency.name)
                for currency
                in self.index_panel_data.currencies]

    def _get_index_levels_choices(self):
        return [(index_level.identity, index_level.name)
                for index_level
                in self.index_panel_data.index_levels]

    def _get_frequencies_choices(self):
        return [(frequency.identity, frequency.name)
                for frequency
                in self.index_panel_data.frequencies]

    def _get_index_suite_groups_choices(self):
        return [(index_suite_group.name, index_suite_group.name)
                for index_suite_group
                in self.index_panel_data.index_suite_groups]

    def _get_index_suites_choices(self):
        suites_without_groups = [
            (suite.identity, suite.name)
            for suite in self.index_panel_data.index_suites
            if suite.group is None]
        grouped_suites = [
            (group.name, [(suite.identity, suite.name)
                          for suite
                          in self.index_panel_data.index_suites
                          if suite.group == group])
            for group
            in self.index_panel_data.index_suite_groups]

        # noinspection PyTypeChecker
        return suites_without_groups + grouped_suites

    def _get_sizes_choices(self):
        return [(size.identity, size.name)
                for size
                in self.index_panel_data.sizes]

    def _get_styles_choices(self):
        return [(style.identity, style.name)
                for style
                in self.index_panel_data.styles]


class MsciStringDataDownloader(InstrumentStringDataDownloader):
    """ Data downloader from app2.msci.com.
    """

    index_data_url = 'https://app2.msci.com/products/service/index/indexmaster/getLevelDataForGraph'
    search_indexes_url = 'https://app2.msci.com/products/service/index/indexmaster/searchIndexes'
    minimal_date_to = datetime.date(1997, 1, 1)

    def __init__(
            self,
            downloader: Downloader,
            download_parameter_values_storage: MsciIndexDownloadParameterValuesStorage):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.downloader = downloader
        self.download_parameter_values_storage = download_parameter_values_storage

        # headers for HTTP
        self.headers: typing.Dict[str, str] = {}

    def adjust_download_instrument_history_parameters(
            self,
            parameters: MsciIndexHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime
    ) -> typing.Tuple[MsciIndexHistoryDownloadParameters, datetime.datetime, datetime.datetime]:

        parameters, moment_from, moment_to = super().adjust_download_instrument_history_parameters(
            parameters, moment_from, moment_to)
        parameters: MsciIndexHistoryDownloadParameters

        if moment_to.date() < self.minimal_date_to:
            moment_to = datetime.datetime.combine(self.minimal_date_to, datetime.time.min, tzinfo=moment_to.tzinfo)

        return parameters, moment_from, moment_to

    def download_instrument_history_string(
            self,
            parameters: MsciIndexHistoryDownloadParameters,
            moment_from: datetime.datetime,
            moment_to: datetime.datetime) -> DownloadStringResult:
        date_from = moment_from.date()
        date_to = moment_to.date()
        if moment_to.time() != datetime.time.min:
            date_to += datetime.timedelta(days=1)

        return self.download_index_history_string(
            parameters.index_code,
            parameters.currency,
            parameters.index_variant,
            date_from,
            date_to)

    def download_instruments_info_string(
            self,
            parameters: MsciIndexesInfoDownloadParameters) -> DownloadStringResult:
        return self.download_indexes_info_string(
            parameters.index_scope,
            parameters.index_market,
            parameters.index_size,
            parameters.index_style,
            parameters.index_suite)

    def download_index_history_string(
            self,
            index_code: str,
            currency: Currency,
            index_variant: IndexLevel,
            date_from: datetime.date,
            date_to: datetime.date) -> DownloadStringResult:
        """ Downloads data for one index as string

        :param index_code: Index code.
        :param currency: Currency
        :param index_variant: Index level.
        :param date_from: Download interval beginning.
        :param date_to: Download interval ending.
        :return: Container with downloaded string.
        """
        params = [
            ('currency_symbol', str(currency.identity)),
            ('index_variant', str(index_variant.identity)),
            ('start_date', date_from.strftime("%Y%m%d")),
            ('end_date', date_to.strftime("%Y%m%d")),
            ('data_frequency', str(self.download_parameter_values_storage.daily_frequency.identity)),
            ('index_codes', str(index_code)),
        ]

        self.downloader.parameters = params
        self.downloader.headers = self.headers

        return self.downloader.download_string(self.index_data_url)

    def download_indexes_info_string(
            self,
            scope: Scopes,
            market: Market,
            size: Size,
            style: Style,
            index_suite: IndexSuite) -> DownloadStringResult:
        """ Downloads the list of all available indexes by specified parameters

        :param scope: Index scope.
        :param market: Index market.
        :param size: Index size.
        :param style: Index style.
        :param index_suite: Index suite.
        :return: Container with downloaded string.
        """
        self.downloader.headers = {}
        self.downloader.parameters = [
            ('index_market', market.identity),
            ('index_scope', scope.value),
            ('index_size', size.identity),
            ('index_style', style.identity),
            ('index_suite', index_suite.identity)
        ]

        return self.downloader.download_string(self.search_indexes_url)


class MsciApiActualityChecker(ApiActualityChecker):
    """ Verifies actuality and accessibility of REST API of app2.msci.com.
    """

    _expected_index_market_id = '16384'  # Developed Markets (DM)
    _expected_index_scope = Scopes.REGIONAL
    _expected_index_size_id = '12'  # Standard (Large+Mid Cap)
    _expected_index_style_id = 'None'  # None
    _expected_index_suite_id = 'C'  # None
    _expected_index_code = '990300'
    _expected_index_name = 'EAFE'
    _expected_currency_id = 'USD'  # USD
    _expected_index_level_id = 'STRD'  # Price
    _expected_start_date = datetime.date(1969, 12, 31)
    _expected_end_date = datetime.date(1997, 1, 1)
    _expected_value = decimal.Decimal(100)

    def __init__(self,
                 parameter_values_storage: MsciIndexDownloadParameterValuesStorage,
                 string_data_downloader: MsciStringDataDownloader,
                 history_json_parser: MsciHistoryJsonParser,
                 index_info_parser: MsciIndexInfoParser):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        self.parameter_values_storage = parameter_values_storage
        self.string_data_downloader = string_data_downloader
        self.history_json_parser = history_json_parser
        self.index_info_parser = index_info_parser

    def check(self):
        self.logger.info("Check actuality via index list")

        market = self.parameter_values_storage.get_dynamic_enum_value_by_key(Market, self._expected_index_market_id)
        if market is None:
            raise CheckApiActualityError(f"Not found market {self._expected_index_market_id!r}")

        size = self.parameter_values_storage.get_dynamic_enum_value_by_key(Size, self._expected_index_size_id)
        if size is None:
            raise CheckApiActualityError(f"Not found size {self._expected_index_size_id!r}")

        style = self.parameter_values_storage.get_dynamic_enum_value_by_key(Style, self._expected_index_style_id)
        if style is None:
            raise CheckApiActualityError(f"Not found style {self._expected_index_style_id!r}")

        suite = self.parameter_values_storage.get_dynamic_enum_value_by_key(IndexSuite, self._expected_index_suite_id)
        if suite is None:
            raise CheckApiActualityError(f"Not found index suite {self._expected_index_suite_id!r}")

        currency = self.parameter_values_storage.get_dynamic_enum_value_by_key(Currency, self._expected_currency_id)
        if currency is None:
            raise CheckApiActualityError(f"Not found currency {self._expected_currency_id!r}")

        index_level = self.parameter_values_storage.get_dynamic_enum_value_by_key(IndexLevel,
                                                                                  self._expected_index_level_id)
        if index_level is None:
            raise CheckApiActualityError(f"Not found index level {self._expected_index_level_id!r}")

        json_data_result = self.string_data_downloader.download_indexes_info_string(
            scope=self._expected_index_scope, market=market, size=size, style=style, index_suite=suite)

        self.logger.debug(f"Got JSON data:\n{json_data_result.downloaded_string}")

        try:
            info_dict = {index_info.msci_index_code: index_info
                         for index_info
                         in self.index_info_parser.parse(json_data_result.downloaded_string)}
        except ParseError as ex:
            json_data_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected index info JSON: {ex.message}") from ex
        except Exception:
            json_data_result.set_correctness(False)
            raise

        if len(info_dict) == 0:
            json_data_result.set_correctness(False)
            raise CheckApiActualityError("Unexpected index info list. No data")

        index_info = info_dict.get(self._expected_index_code, None)
        if index_info is None:
            json_data_result.set_correctness(False)
            raise CheckApiActualityError(f"Not found index with code {self._expected_index_code!r}")
        if index_info.index_name != self._expected_index_name:
            json_data_result.set_correctness(False)
            raise CheckApiActualityError(f"Index with code {self._expected_index_code!r} changed name "
                                         f"from {self._expected_index_name} to {index_info.index_name}")

        # here we know that check via index list was successful

        self.logger.info(f"Check actuality via index {self._expected_index_name!r}")

        json_history_result = self.string_data_downloader.download_index_history_string(
            index_code=self._expected_index_code,
            currency=currency,
            index_variant=index_level,
            date_from=self._expected_start_date,
            date_to=self._expected_end_date
        )

        try:
            history = list(self.history_json_parser.parse(json_history_result.downloaded_string, tzinfo=None))
        except ParseError as ex:
            json_history_result.set_correctness(False)
            raise CheckApiActualityError(f"Unexpected index history JSON: {ex.message}") from ex
        except Exception:
            json_history_result.set_correctness(False)
            raise

        if not history:
            json_history_result.set_correctness(False)
            raise CheckApiActualityError("History is empty")

        history.sort(key=lambda v: v.calc_date)

        first_value = history[0]  # for actuality checking one record is enough

        if first_value.msci_index_code != self._expected_index_code:
            json_history_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected index code. {first_value.msci_index_code!r} != {self._expected_index_code!r}")

        if first_value.calc_date != self._expected_start_date:
            json_history_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected first date. {first_value.calc_date} != {self._expected_start_date}")

        if first_value.level_eod != self._expected_value:
            json_history_result.set_correctness(False)
            raise CheckApiActualityError(
                f"Unexpected first value. {first_value.level_eod} != {self._expected_value}")

        self.logger.info("Actuality check was successful")


class MsciIndexExporterFactory(InstrumentExporterFactory):
    """ Factory class for create instances of MSCI data exporter.
    """
    name: str = 'MSCI index data exporter. Version 2021.'
    provider_site: str = 'https://www.msci.com/'
    api_url: str = 'https://app2.msci.com/products/index-data-search/'

    def __init__(self):
        self._dynamic_enum_type_manager = None
        self._download_parameters_factory = None

    def create_history_values_exporter(self, downloader: Downloader) -> InstrumentHistoryValuesExporter:
        index_panel_data_json_parser = MsciIndexPanelDataJsonParser()
        parameter_values_storage = MsciIndexDownloadParameterValuesStorage(downloader, index_panel_data_json_parser)
        string_data_downloader = MsciStringDataDownloader(downloader, parameter_values_storage)
        history_values_parser = MsciHistoryJsonParser(parameter_values_storage)

        return GenericInstrumentHistoryValuesExporter(string_data_downloader, history_values_parser)

    def create_info_exporter(self, downloader: Downloader) -> InstrumentsInfoExporter:
        index_panel_data_json_parser = MsciIndexPanelDataJsonParser()
        parameter_values_storage = MsciIndexDownloadParameterValuesStorage(downloader, index_panel_data_json_parser)
        string_data_downloader = MsciStringDataDownloader(downloader, parameter_values_storage)
        info_parser = MsciIndexInfoParser()

        return GenericInstrumentsInfoExporter(string_data_downloader, info_parser)

    def create_download_parameter_values_storage(self, downloader: Downloader) -> DownloadParameterValuesStorage:
        index_panel_data_json_parser = MsciIndexPanelDataJsonParser()
        return MsciIndexDownloadParameterValuesStorage(downloader, index_panel_data_json_parser)

    def create_api_actuality_checker(self, downloader: Downloader) -> MsciApiActualityChecker:
        index_panel_data_json_parser = MsciIndexPanelDataJsonParser()
        parameter_values_storage = MsciIndexDownloadParameterValuesStorage(downloader, index_panel_data_json_parser)
        string_data_downloader = MsciStringDataDownloader(downloader, parameter_values_storage)
        history_values_parser = MsciHistoryJsonParser(parameter_values_storage)
        info_parser = MsciIndexInfoParser()

        return MsciApiActualityChecker(
            parameter_values_storage,
            string_data_downloader,
            history_values_parser,
            info_parser)

    @property
    def dynamic_enum_type_manager(self) -> MsciDynamicEnumTypeManager:
        if self._dynamic_enum_type_manager is None:
            self._dynamic_enum_type_manager = MsciDynamicEnumTypeManager()

        return self._dynamic_enum_type_manager

    @property
    def download_parameters_factory(self) -> MsciDownloadParametersFactory:
        if self._download_parameters_factory is None:
            self._download_parameters_factory = MsciDownloadParametersFactory()

        return self._download_parameters_factory



#!/usr/bin/python
# -*- coding: utf-8 -*-

""" Simplified utility for retry strategy with exponential backoff.
"""

import logging
import time
import typing

logging.getLogger().addHandler(logging.NullHandler())


def wait_and_retry(action: typing.Callable[[], typing.Any],
                   transient_validator: typing.Callable[[Exception, int], bool],
                   max_retry_count: int = 10,
                   first_sleep_duration: float = 1.0,
                   sleep_duration_factor: float = 2.0):
    """ Implements a policy that will wait and retry `max_retry_count` times calling `action`
    on each retry with the raised exception and the current sleep duration.

    The duration of wait (sleep) between retries calculates as
    ``first_sleep_duration * (sleep_duration_factor ^ retry_attempt)``,
    where `retry_attempt` is zero based count of `action` calls.

    :param action: The action to call on each try.
    :param transient_validator: Determine if the exception is transient, i.e. available for retry.
    :param max_retry_count: The max retry count (not including first call).
    :param first_sleep_duration: Duration in seconds to wait for a first retry attempt
      (right after first failed `action`).
    :param sleep_duration_factor: Factor for the duration to wait before next (second, etc.) retries.
    :return:
    """
    logger = logging.getLogger(__name__ + '.wait_and_retry')

    if not callable(action):
        logger.error(f"Action '{action}' is not callable")
        raise ValueError(f"Action '{action}' is not callable")

    if not callable(transient_validator):
        logger.error(f"Transient validator '{transient_validator}' is not callable")
        raise ValueError(f"Transient validator '{transient_validator}' is not callable")

    min_max_retry_count = 0
    if max_retry_count < min_max_retry_count:
        logger.debug(f"max_retry_count adjusted from {max_retry_count} to {min_max_retry_count}")
        max_retry_count = min_max_retry_count
    del min_max_retry_count

    min_first_sleep_duration = 0.0
    if first_sleep_duration < min_first_sleep_duration:
        logger.debug(f"first_sleep_duration adjusted from {first_sleep_duration} to {min_first_sleep_duration}")
        first_sleep_duration = min_first_sleep_duration
    del min_first_sleep_duration

    adjusted_sleep_duration_factor = 1.0
    if sleep_duration_factor < 0.0:
        logger.debug(f"sleep_duration_factor adjusted from {sleep_duration_factor} to {adjusted_sleep_duration_factor}")
        sleep_duration_factor = adjusted_sleep_duration_factor
    del adjusted_sleep_duration_factor

    pause_in_seconds = first_sleep_duration

    retry_attempt = 0

    while retry_attempt <= max_retry_count:  # always do at least one attempt # pragma: no branch

        logger.log(retry_attempt > 0 and logging.INFO or logging.DEBUG, f"Attempt #{retry_attempt}...")

        try:
            action()

        except Exception as ex:  # pylint: disable=broad-except
            logger.exception("Attempt failed")

            if retry_attempt >= max_retry_count or not transient_validator(ex, retry_attempt + 1):
                raise

            pause_in_seconds *= sleep_duration_factor

        else:
            # success
            logger.info("Attempt succeeded")

            return

        if pause_in_seconds > 0.0:
            logger.debug(f"Wait for {pause_in_seconds} seconds before next attempt...")
            time.sleep(pause_in_seconds)

        retry_attempt += 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import datetime
import decimal

from sane_finances.sources.base import CheckApiActualityError, ParseError, InstrumentExporterFactory
from sane_finances.sources.msci.v2021.exporters import (
    MsciDynamicEnumTypeManager, MsciIndexDownloadParameterValuesStorage,
    MsciStringDataDownloader, MsciApiActualityChecker, MsciIndexExporterFactory)
from sane_finances.sources.msci.v2021.meta import (
    MsciIndexHistoryDownloadParameters, MsciIndexesInfoDownloadParameters,
    IndexValue, IndexInfo, IndexPanelData, Frequency, Market, IndexSuite, IndexSuiteGroup,
    Style, Size, Scopes, IndexLevel, Currency)

from .fakes import (
    FakeDownloader, FakeMsciIndexPanelDataJsonParser,
    FakeIndexInfoParser, FakeMsciHistoryJsonParser, FakeMsciStringDataDownloader,
    FakeMsciIndexDownloadParameterValuesStorage)
from .common import CommonTestCases


class TestMsciDynamicEnumTypeManager(unittest.TestCase):

    def test_is_dynamic_enum_type_Success(self):
        manager = MsciDynamicEnumTypeManager()
        all_types = manager.get_all_managed_types()

        for dynamic_enum_type in all_types:
            self.assertTrue(manager.is_dynamic_enum_type(dynamic_enum_type))

    def test_is_dynamic_enum_type_FalseWithWrongType(self):
        manager = MsciDynamicEnumTypeManager()

        # noinspection PyTypeChecker
        self.assertFalse(manager.is_dynamic_enum_type(None))

    def test_get_dynamic_enum_key_Success(self):
        manager = MsciDynamicEnumTypeManager()
        instance = Market(identity='ID', name='NAME')

        self.assertEqual(manager.get_dynamic_enum_key(instance), instance.identity)

    def test_get_dynamic_enum_key_NoneWithWrongType(self):
        manager = MsciDynamicEnumTypeManager()

        self.assertIsNone(manager.get_dynamic_enum_key(None))


class TestMsciIndexDownloadParameterValuesStorage(unittest.TestCase):

    def setUp(self):
        self.daily_frequency = Frequency(identity='DAILY', name='Daily')
        self.monthly_frequency = Frequency(identity='END_OF_MONTH', name='Monthly')
        index_suite_group = IndexSuiteGroup(name='ID')
        self.fake_index_panel_data = IndexPanelData(
            markets=(Market(identity='ID', name='NAME'),),
            currencies=(Currency(identity='ID', name='NAME'),),
            index_levels=(IndexLevel(identity='ID', name='NAME'),),
            frequencies=(self.daily_frequency, self.monthly_frequency, Frequency(identity='ID', name='NAME')),
            index_suite_groups=(index_suite_group,),
            index_suites=(IndexSuite(identity='ID', name='NAME', group=index_suite_group),),
            sizes=(Size(identity='ID', name='NAME'),),
            styles=(Style(identity='ID', name='NAME'),),
            daily_frequency=self.daily_frequency,
            monthly_frequency=self.monthly_frequency
        )

        self.downloader = FakeDownloader(None)
        self.index_panel_data_json_parser = FakeMsciIndexPanelDataJsonParser(self.fake_index_panel_data)
        self.storage = MsciIndexDownloadParameterValuesStorage(
            self.downloader,
            self.index_panel_data_json_parser)

    def test_reload_Success(self):
        self.storage.reload()

        self.assertGreaterEqual(len(self.downloader.download_string_results), 1)
        self.assertTrue(all(result.is_correct
                            for result
                            in self.downloader.download_string_results))

    def test_reload_LastDownloadStringResultIsNotCorrectWhenError(self):
        self.index_panel_data_json_parser.parse_exception = ParseError('Error')

        with self.assertRaises(Exception):
            self.storage.reload()

        self.assertGreaterEqual(len(self.downloader.download_string_results), 1)
        self.assertIs(self.downloader.download_string_results[-1].is_correct, False)

    def test_daily_frequency_Success(self):
        self.assertEqual(self.storage.daily_frequency, self.daily_frequency)

    def test_monthly_frequency_Success(self):
        self.assertEqual(self.storage.monthly_frequency, self.monthly_frequency)

    def test_get_dynamic_enum_value_by_key_Success(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            value = self.storage.get_dynamic_enum_value_by_key(dynamic_enum_type, 'ID')

            self.assertIsNotNone(value)

        # noinspection PyTypeChecker
        value = self.storage.get_dynamic_enum_value_by_key(None, 'ID')

        self.assertIsNone(value)

    def test_get_dynamic_enum_value_by_choice_Success(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            value = self.storage.get_dynamic_enum_value_by_choice(dynamic_enum_type, 'ID')

            self.assertIsNotNone(value)

        # noinspection PyTypeChecker
        value = self.storage.get_dynamic_enum_value_by_choice(None, 'ID')

        self.assertIsNone(value)

    def test_get_all_parameter_values_for_Success(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            value = self.storage.get_all_parameter_values_for(dynamic_enum_type)

            self.assertIsNotNone(value)

        # noinspection PyTypeChecker
        value = self.storage.get_all_parameter_values_for(None)

        self.assertIsNone(value)

    def test_get_parameter_type_choices_Success(self):
        all_types = self.storage.get_all_managed_types()
        for dynamic_enum_type in all_types:
            choices = self.storage.get_parameter_type_choices(dynamic_enum_type)

            self.assertIsNotNone(choices)

        # noinspection PyTypeChecker
        choices = self.storage.get_parameter_type_choices(None)

        self.assertIsNone(choices)


class TestMsciStringDataDownloader(unittest.TestCase):

    def setUp(self):
        self.fake_data = 'data'

        daily_frequency = Frequency(identity='FREQ ID', name='FREQ NAME')
        # noinspection PyTypeChecker
        self.index_panel_data = IndexPanelData(
            markets=(Market(identity='ID', name='NAME'),),
            currencies=(Currency(identity='ID', name='NAME'),),
            index_levels=(IndexLevel(identity='ID', name='NAME'),),
            frequencies=(),
            index_suite_groups=(),
            index_suites=(IndexSuite(identity='ID', name='NAME'),),
            sizes=(Size(identity='ID', name='NAME'),),
            styles=(Style(identity='ID', name='NAME'),),
            daily_frequency=daily_frequency,
            monthly_frequency=None
        )

        self.string_data_downloader = MsciStringDataDownloader(
            FakeDownloader(self.fake_data),
            FakeMsciIndexDownloadParameterValuesStorage(self.index_panel_data)
        )
        self.history_download_params = MsciIndexHistoryDownloadParameters(
            index_code='WHATEVER',
            currency=self.index_panel_data.currencies[0],
            index_variant=self.index_panel_data.index_levels[0])

    def test_adjust_download_instrument_history_parameters_Success(self):
        date_to = datetime.datetime.combine(
            self.string_data_downloader.minimal_date_to - datetime.timedelta(days=1),
            datetime.time.min)  # date before minimum allowed date to
        date_from = date_to - datetime.timedelta(days=1)

        params, moment_from, moment_to = self.string_data_downloader.adjust_download_instrument_history_parameters(
            parameters=self.history_download_params,
            moment_from=date_from,
            moment_to=date_to)

        self.assertEqual(moment_to.date(), self.string_data_downloader.minimal_date_to)
        self.assertLessEqual(moment_from, moment_to)
        self.assertIsNotNone(params)

    def test_download_instrument_history_string_Success(self):
        moment_from = datetime.datetime(2010, 1, 1, 12)  # has hours
        moment_to = moment_from + datetime.timedelta(days=1)

        result = self.string_data_downloader.download_instrument_history_string(
            self.history_download_params,
            moment_from,
            moment_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_index_history_string_Success(self):
        date_from = datetime.date(2010, 1, 1)
        date_to = date_from + datetime.timedelta(days=1)

        result = self.string_data_downloader.download_index_history_string(
            'WHATEVER',
            self.index_panel_data.currencies[0],
            self.index_panel_data.index_levels[0],
            date_from,
            date_to)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_instruments_info_string_Success(self):
        params = MsciIndexesInfoDownloadParameters.safe_create(
            index_scope=Scopes.REGIONAL,
            index_market=self.index_panel_data.markets[0],
            index_size=self.index_panel_data.sizes[0],
            index_style=self.index_panel_data.styles[0],
            index_suite=self.index_panel_data.index_suites[0])

        result = self.string_data_downloader.download_instruments_info_string(params)

        self.assertEqual(result.downloaded_string, self.fake_data)

    def test_download_indexes_info_string_Success(self):
        result = self.string_data_downloader.download_indexes_info_string(
            Scopes.REGIONAL,
            self.index_panel_data.markets[0],
            self.index_panel_data.sizes[0],
            self.index_panel_data.styles[0],
            self.index_panel_data.index_suites[0])

        self.assertEqual(result.downloaded_string, self.fake_data)


class TestMsciAPIActualityChecker(unittest.TestCase):

    def setUp(self):
        daily_frequency = Frequency(identity='DAY ID', name='FREQ NAME')
        monthly_frequency = Frequency(identity='MONTH ID', name='FREQ NAME')
        self.index_panel_data = IndexPanelData(
            markets=(Market(identity=MsciApiActualityChecker._expected_index_market_id, name='NAME'),),
            currencies=(Currency(identity=MsciApiActualityChecker._expected_currency_id, name='NAME'),),
            index_levels=(IndexLevel(identity=MsciApiActualityChecker._expected_index_level_id, name='NAME'),),
            frequencies=(daily_frequency,),
            index_suite_groups=(),
            index_suites=(IndexSuite(identity=MsciApiActualityChecker._expected_index_suite_id, name='NAME'),),
            sizes=(Size(identity=MsciApiActualityChecker._expected_index_size_id, name='NAME'),),
            styles=(Style(identity=MsciApiActualityChecker._expected_index_style_id, name='NAME'),),
            daily_frequency=daily_frequency,
            monthly_frequency=monthly_frequency
        )

        self.success_info_data = [IndexInfo(
            msci_index_code=MsciApiActualityChecker._expected_index_code,
            index_name=MsciApiActualityChecker._expected_index_name)
        ]
        self.index_info_parser = FakeIndexInfoParser(self.success_info_data)

        self.success_history_data = [IndexValue(
            calc_date=MsciApiActualityChecker._expected_start_date,
            level_eod=MsciApiActualityChecker._expected_value,
            msci_index_code=MsciApiActualityChecker._expected_index_code,
            index_variant_type=self.index_panel_data.index_levels[0],
            currency=self.index_panel_data.currencies[0]
        )]
        self.history_parser = FakeMsciHistoryJsonParser(self.success_history_data)
        self.string_data_downloader = FakeMsciStringDataDownloader(None, None)

    def get_checker(self):
        return MsciApiActualityChecker(
            FakeMsciIndexDownloadParameterValuesStorage(self.index_panel_data),
            self.string_data_downloader,
            self.history_parser,
            self.index_info_parser)

    def test_check_Success(self):
        checker = self.get_checker()
        checker.check()

        # check that there is no incorrect downloaded strings
        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.string_data_downloader.download_instruments_info_string_results))
        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertFalse(any(result.is_correct is False
                             for result
                             in self.string_data_downloader.download_instrument_history_string_results))

    def test_check_raisesWhenInfoParseError(self):
        # corrupt data
        self.index_info_parser.parse_exception = ParseError('Error')
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenInfoUnknownError(self):
        # corrupt data
        self.index_info_parser.parse_exception = Exception('Error')
        checker = self.get_checker()

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenNoInfo(self):
        # corrupt data
        self.index_info_parser.fake_data = []  # No data
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instruments_info_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instruments_info_string_results[-1].is_correct, False)

    def test_check_raisesWhenHistoryParseError(self):
        # corrupt data
        self.history_parser.parse_exception = ParseError('Error')
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenHistoryUnknownError(self):
        # corrupt data
        self.history_parser.parse_exception = Exception('Error')
        checker = self.get_checker()

        with self.assertRaises(Exception):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenNoHistory(self):
        # corrupt data
        self.history_parser.fake_data = []  # No data
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

        self.assertGreaterEqual(len(self.string_data_downloader.download_instrument_history_string_results), 1)
        self.assertIs(self.string_data_downloader.download_instrument_history_string_results[-1].is_correct, False)

    def test_check_raisesWhenWrongFirstDate(self):
        # corrupt data
        self.success_history_data[0].calc_date = datetime.date.max
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongFirstValue(self):
        # corrupt data
        self.success_history_data[0].level_eod = decimal.Decimal(-1)
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongFirstCode(self):
        # corrupt data
        self.success_history_data[0].msci_index_code = 'WRONG'
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongMarket(self):
        # corrupt data
        self.index_panel_data = self.index_panel_data._replace(
            markets=(Market(identity='WRONG', name='NAME'),))

        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongCurrency(self):
        # corrupt data
        self.index_panel_data = self.index_panel_data._replace(
            currencies=(Currency(identity='WRONG', name='NAME'),))

        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongIndexLevel(self):
        # corrupt data
        self.index_panel_data = self.index_panel_data._replace(
            index_levels=(IndexLevel(identity='WRONG', name='NAME'),))

        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongIndexSuite(self):
        # corrupt data
        self.index_panel_data = self.index_panel_data._replace(
            index_suites=(IndexSuite(identity='WRONG', name='NAME'),))

        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongSize(self):
        # corrupt data
        self.index_panel_data = self.index_panel_data._replace(
            sizes=(Size(identity='WRONG', name='NAME'),))

        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenWrongStyle(self):
        # corrupt data
        self.index_panel_data = self.index_panel_data._replace(
            styles=(Style(identity='WRONG', name='NAME'),))

        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenIncorrectInfoJson(self):
        self.index_info_parser.parse_exception = ParseError('Fake error')
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenIncorrectHistoryJson(self):
        self.history_parser.parse_exception = ParseError('Fake error')
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenUnknownIndexCode(self):
        self.success_info_data[0] = IndexInfo(
            msci_index_code='WRONG',
            index_name=MsciApiActualityChecker._expected_index_name)
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()

    def test_check_raisesWhenUnknownIndexName(self):
        self.success_info_data[0] = IndexInfo(
            msci_index_code=MsciApiActualityChecker._expected_index_code,
            index_name='WRONG')
        checker = self.get_checker()

        with self.assertRaises(CheckApiActualityError):
            checker.check()


class TestMsciIndexExporterFactory(CommonTestCases.CommonIndexExporterFactoryTests):

    def get_exporter_factory(self) -> InstrumentExporterFactory:
        return MsciIndexExporterFactory()

    def is_dynamic_enum_type_manager_singleton(self):
        return True

    def is_download_parameters_factory_singleton(self):
        return True



""" Sources from Central Bank of Russia.
"""
from . import v2016
from ..generic import InstrumentExporterRegistry, register_instrument_history_values_exporter

__all__ = ['v2016']

register_instrument_history_values_exporter(
    InstrumentExporterRegistry(
        name=v2016.exporters.CbrCurrencyRatesExporterFactory.name,
        provider_site=v2016.exporters.CbrCurrencyRatesExporterFactory.provider_site,
        api_url=v2016.exporters.CbrCurrencyRatesExporterFactory.api_url,
        factory=v2016.exporters.CbrCurrencyRatesExporterFactory()
    ))



#!/usr/bin/python
# -*- coding: utf-8 -*-

import unittest

from sane_finances.retry_policy import wait_and_retry


class TestWaitAndRetry(unittest.TestCase):

    def setUp(self):
        self.alwaysRaisingActionCount = 0
        self.properRaisingActionCount = 0

    def alwaysRaisingAction(self, x):
        self.alwaysRaisingActionCount += 1
        raise x

    def properRaisingAction(self, x, num_of_raised):
        """
        - *x* - Exception to be raised
        - *num_of_raised* - Number of calls (not only retries) when exception *x* raises
        """
        self.properRaisingActionCount += 1
        if self.properRaisingActionCount <= num_of_raised:
            raise x

    def test_raisedForNotValidAction(self):
        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            wait_and_retry(None, lambda: True)

    def test_raisedForNotValidTransientValidator(self):
        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            wait_and_retry(lambda: None, None)

    def test_actionCallsProperCountOfTimesWIthSleep(self):
        retry_count = 2
        estimated_action_calls = retry_count + 1

        with self.assertRaises(StopIteration):
            wait_and_retry(
                action=lambda: self.alwaysRaisingAction(StopIteration),
                transient_validator=lambda x, n: True,
                max_retry_count=retry_count,
                sleep_duration_factor=0.01,
                first_sleep_duration=0.01  # sleep a little
            )

        self.assertEqual(
            self.alwaysRaisingActionCount,
            estimated_action_calls,
            "Action called not proper num of times"
        )

    def test_actionCallsProperCountOfTimesAndAdjustParams(self):
        estimated_action_calls = 1

        with self.assertRaises(StopIteration):
            wait_and_retry(
                action=lambda: self.alwaysRaisingAction(StopIteration),
                transient_validator=lambda x, n: True,
                max_retry_count=-1,
                sleep_duration_factor=-1.0,
                first_sleep_duration=-1.0
            )

        self.assertEqual(
            self.alwaysRaisingActionCount,
            estimated_action_calls,
            "Action called not proper num of times"
        )

    def test_actionCallsProperCountOfTimesNeverRetry(self):
        retry_count = 0
        estimated_action_calls = retry_count + 1

        with self.assertRaises(StopIteration):
            wait_and_retry(
                action=lambda: self.alwaysRaisingAction(StopIteration),
                transient_validator=lambda x, n: True,
                max_retry_count=retry_count,
                first_sleep_duration=0.0  # do not sleep
            )

        self.assertEqual(
            self.alwaysRaisingActionCount,
            estimated_action_calls,
            "Action called not proper num of times"
        )

    def test_actionCallsProperCountOfTimes(self):
        retry_count = 5
        estimated_action_calls = retry_count + 1

        with self.assertRaises(StopIteration):
            wait_and_retry(
                action=lambda: self.alwaysRaisingAction(StopIteration),
                transient_validator=lambda x, n: True,
                max_retry_count=retry_count,
                first_sleep_duration=0.0  # do not sleep
            )

        self.assertEqual(
            self.alwaysRaisingActionCount,
            estimated_action_calls,
            "Action called not proper num of times"
        )

    def test_actionCallsProperCountOfTimesIfTransient(self):
        transient_retry_count = 5  # num of retries when exception is supposed to be transient;
        # after that every exception is critical

        retry_count = transient_retry_count * 2  # retry much more than *transient_retry_count* admits
        estimated_action_calls = transient_retry_count + 1

        self.assertGreater(retry_count, transient_retry_count, "Wrong test setting")

        with self.assertRaises(StopIteration):
            wait_and_retry(
                action=lambda: self.alwaysRaisingAction(StopIteration),
                transient_validator=lambda x, n: n <= transient_retry_count,
                max_retry_count=retry_count,
                first_sleep_duration=0.0  # do not sleep
            )

        self.assertEqual(
            self.alwaysRaisingActionCount,
            estimated_action_calls,
            "Action called not proper num of times for transient exception"
        )

    def test_successAfterFailedActions(self):
        failed_calls = 5
        retry_count = failed_calls + 1
        estimated_action_calls = failed_calls + 1

        self.assertGreater(retry_count, failed_calls, "Wrong test setting")

        wait_and_retry(
            action=lambda: self.properRaisingAction(StopIteration, failed_calls),
            transient_validator=lambda x, n: True,
            max_retry_count=retry_count,
            first_sleep_duration=0.0  # do not sleep
        )

        self.assertEqual(
            self.properRaisingActionCount,
            estimated_action_calls,
            "Action called not proper num of times before success"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Utilities for parse data from solactive.com
"""

import json
import decimal
import datetime
import logging
import re
import typing

from .meta import IndexInfo, IndexValue, FieldNames
from ...base import InstrumentValuesHistoryParser, InstrumentInfoParser, ParseError

logging.getLogger().addHandler(logging.NullHandler())


class SolactiveJsonParser(InstrumentValuesHistoryParser):
    """ Parser for history data of index from JSON string.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def parse(  # pylint: disable=arguments-renamed
            self,
            raw_json_text: str,
            tzinfo: typing.Optional[datetime.timezone]
    ) -> typing.Iterable[IndexValue]:

        try:
            raw_data = json.loads(raw_json_text)
        except json.decoder.JSONDecodeError as ex:
            raise ParseError(ex.msg) from ex

        if not isinstance(raw_data, list):
            # can be Inf etc., but we accept only []
            raise ParseError("Wrong JSON format. Top level is not list.")

        if len(raw_data) == 0:
            # empty list make no sense: there always must be history
            raise ParseError("Empty JSON")

        first_date = datetime.datetime(1970, 1, 1, tzinfo=tzinfo)

        for json_item in raw_data:
            if not isinstance(json_item, dict):
                raise ParseError("Wrong JSON format. Item level is not dict.")

            for field_name in (FieldNames.INDEX_ID.value, FieldNames.TIMESTAMP.value, FieldNames.VALUE.value):
                if field_name not in json_item:
                    raise ParseError(f"Wrong JSON format. Has no '{field_name}' field.")

            moment = first_date + datetime.timedelta(milliseconds=json_item[FieldNames.TIMESTAMP.value])

            # convert from float to Decimal
            str_value = "%.4f" % json_item[FieldNames.VALUE.value]
            value = decimal.Decimal(str_value)

            self.logger.debug(f"Converted value from {json_item[FieldNames.VALUE.value]} to {value}")

            yield IndexValue(index_id=json_item[FieldNames.INDEX_ID.value], moment=moment, value=value)


class SolactiveIndexInfoParser(InstrumentInfoParser):
    """ Parser for indexes info list from HTML.
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def parse(self, html: str) -> typing.Iterable[IndexInfo]:  # pylint: disable=arguments-renamed
        table_pattern = re.compile(
            r"<tbody.*?>(?P<table>.*?)</tbody.*?>",
            re.IGNORECASE | re.MULTILINE | re.DOTALL
        )
        row_pattern = re.compile(
            r".*?<tr.*?>(?P<row>.*?)</tr.*?>.*?",
            re.IGNORECASE | re.MULTILINE | re.DOTALL
        )
        cell_pattern = re.compile(
            r"<td.*?class=\"(?P<class_name>.*?)\".*?>(?P<cell>.*?)</td.*?>",
            re.IGNORECASE | re.MULTILINE | re.DOTALL
        )
        name_cell_pattern = re.compile(
            r".*?<a.*?>(?P<name>.*?)</a.*?>.*?",
            re.IGNORECASE | re.MULTILINE | re.DOTALL
        )

        has_any = False
        for m_table in table_pattern.finditer(html):
            for m_row in row_pattern.finditer(m_table.group("table")):
                row = m_row.group("row")

                name = None
                isin = None

                for m_cell in cell_pattern.finditer(row):
                    class_name = m_cell.group("class_name")
                    cell = m_cell.group("cell")

                    if class_name == "name":
                        m_name = name_cell_pattern.match(cell)
                        if m_name:
                            name = m_name.group("name")

                    elif class_name == "isin":
                        isin = cell

                if name is None:
                    self.logger.error(f"Index name not found in {row!r}")
                    raise ParseError("Unexpected HTML format. Index name not found")

                if isin is None:
                    self.logger.error(f"Index ISIN not found in {row!r}")
                    raise ParseError("Unexpected HTML format. Index ISIN not found")

                has_any = True
                yield IndexInfo(name=name, isin=isin)

        if not has_any:
            self.logger.error(f"No data found in HTML:\n{html}")
            raise ParseError("Unexpected HTML format. No data found")



""" Library for download and parse financial data from various sources.
"""

__author__ = "Stanley Sane"
__copyright__ = "Copyright (c) 2018-2021, Sane Development"
__license__ = "The BSD 3-Clause License"

__version__ = '1.0'
